# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining mock data and classes for testing the runtime module.

"""
from typing import Any, Optional
from unittest.mock import MagicMock

from qbraid_core.services.quantum.exceptions import QuantumServiceRequestError

from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus

DEVICE_DATA_QIR = {
    "numberQubits": 64,
    "pendingJobs": 0,
    "qbraid_id": "qbraid_qir_simulator",
    "name": "QIR sparse simulator",
    "provider": "qBraid",
    "paradigm": "gate-based",
    "type": "SIMULATOR",
    "vendor": "qBraid",
    "runInputTypes": ["pyqir"],
    "status": "ONLINE",
    "isAvailable": True,
    "processorType": "State vector",
    "noiseModels": ["ideal"],
    "pricing": {"perTask": 0.005, "perShot": 0, "perMinute": 0.075},
}

DEVICE_DATA_QUERA_QASM = {
    "numberQubits": 30,
    "pendingJobs": 0,
    "qbraid_id": "quera_qasm_simulator",
    "name": "Noisey QASM simulator",
    "provider": "QuEra",
    "paradigm": "gate-based",
    "type": "SIMULATOR",
    "vendor": "qBraid",
    "runInputTypes": ["qasm2"],
    "status": "ONLINE",
    "isAvailable": True,
    "processorType": "State vector",
    "noiseModels": ["quera_lqs_backend"],
    "pricing": {"perTask": 0, "perShot": 0, "perMinute": 0},
}

DEVICE_DATA_NEC = {
    "qbraid_id": "nec_vector_annealer",
    "name": "NEC Vector Annealer",
    "provider": "NEC",
    "paradigm": "annealing",
    "type": "SIMULATOR",
    "vendor": "qbraid",
    "status": "ONLINE",
    "isAvailable": True,
    "numberQubits": 0,
    "runInputTypes": ["qubo"],
    "pricing": {"perMinute": 0, "perShot": 0.00145, "perTask": 0.3},
}

DEVICE_DATA_AQUILA = {
    "numberQubits": 256,
    "pendingJobs": 9,
    "qbraid_id": "quera_aquila",
    "name": "Aquila",
    "provider": "QuEra",
    "paradigm": "AHS",
    "type": "QPU",
    "vendor": "AWS",
    "runInputTypes": ["braket_ahs"],
    "status": "OFFLINE",
    "isAvailable": False,
    "architecture": "neutral atom",
    "pricing": {"perTask": 0.3, "perShot": 0.01, "perMinute": 0},
}

REDUNDANT_JOB_DATA = {
    "timeStamps": {
        "createdAt": "2024-05-23T01:39:11.288Z",
        "endedAt": "2024-05-23T01:39:11.304Z",
        "executionDuration": 16,
    },
    "status": "COMPLETED",
}

JOB_DATA_QIR = {
    "qbraidJobId": "qbraid_qir_simulator-jovyan-qjob-1234567890",
    "queuePosition": None,
    "queueDepth": None,
    "shots": 10,
    "circuitNumQubits": 5,
    "qbraidDeviceId": "qbraid_qir_simulator",
    "vendor": "qbraid",
    "provider": "qbraid",
    "tags": {},
    "experimentType": "gate_model",
    **REDUNDANT_JOB_DATA,
}

JOB_DATA_QUERA_QASM = {
    "qbraidJobId": "quera_qasm_simulator-jovyan-qjob-1234567890",
    "queuePosition": None,
    "queueDepth": None,
    "shots": 10,
    "circuitNumQubits": 5,
    "qbraidDeviceId": "quera_qasm_simulator",
    "vendor": "qbraid",
    "provider": "quera",
    "tags": {},
    "experimentType": "gate_model",
    **REDUNDANT_JOB_DATA,
}

JOB_DATA_NEC = {
    "qbraidJobId": "nec_vector_annealer-jovyan-qjob-1234567890",
    "queuePosition": None,
    "queueDepth": None,
    "qubo": {("x1", "x1"): 3.0, ("x1", "x2"): 2.0},
    "offset": 0.0,
    "qbraidDeviceId": "nec_vector_annealer",
    "vendor": "qbraid",
    "provider": "nec",
    "tags": {},
    "experimentType": "annealing",
    **REDUNDANT_JOB_DATA,
}

JOB_DATA_AQUILA = {
    "qbraidJobId": "quera_aquila-jovyan-qjob-1234567890",
    "queuePosition": None,
    "queueDepth": None,
    "shots": 10,
    "qbraidDeviceId": "quera_aquila",
    "atomRegister": [
        [0.00000285, 0.00000688],
        [0.00000688, 0.00000285],
        [0.00000688, -0.00000285],
        [0.00000285, -0.00000688],
        [-0.00000285, -0.00000688],
        [-0.00000688, -0.00000285],
        [-0.00000688, 0.00000285],
        [-0.00000285, 0.00000688],
    ],
    "filling": [1, 1, 1, 1, 1, 1, 1, 1],
    "numAtoms": 8,
    "vendor": "aws",
    "provider": "quera",
    "tags": {},
    "experimentType": "ahs",
    **REDUNDANT_JOB_DATA,
}

RESULTS_DATA_QIR = {
    "measurements": [
        [0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1],
    ],
    "measurementCounts": {"11111": 4, "00000": 6},
    "runnerVersion": "0.7.4",
    "runnerSeed": None,
    **REDUNDANT_JOB_DATA,
}

RESULTS_DATA_QUERA_QASM = {
    "measurementCounts": {"11111": 4, "00000": 6},
    "backend": "cirq-gpu",
    "quera_simulation_result": {
        "flair_visual_version": "0.5.3",
        "counts": {"0": 6, "31": 4},
        "logs": (
            ",atom_id,block_id,action_type,time,duration\n"
            "0,0,TrapSLM,0,0\n"
            "0,0,TrapAOD,0,31.024984394500784\n"
            "0,0,DropAOD,31.024984394500784,0"
        ),
        "atom_animation_state": {
            "block_durations": [],
            "gate_events": [],
            "qpu_fov": {"xmin": -6, "xmax": 218, "ymin": -5, "ymax": 66.5},
            "atoms": [],
            "slm_zone": [],
            "aod_moves": [],
        },
        "noise_model": {"all_qubits": [0, 1, 2, 3, 4], "gate_events": []},
    },
    **REDUNDANT_JOB_DATA,
}

RESULTS_DATA_NEC = {
    "solutions": [
        {
            "spin": {" x1": 0, " x2": 0, "x1": 0},
            "energy": 0,
            "time": 0.006517000030726194,
            "constraint": True,
            "memory_usage": 1.189453125,
        }
    ],
    "solutionCount": 1,
    **REDUNDANT_JOB_DATA,
}

RESULTS_DATA_AQUILA = {
    "measurementCounts": {
        "grggrggr": 1,
        "grgrgerg": 1,
        "rgrgrgrg": 1,
        "grgrgrgr": 3,
        "ggrggrgr": 1,
        "ggrgggrr": 1,
        "rgrggrgg": 1,
        "rggrgreg": 1,
    },
    "measurements": [
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 0, 1, 1, 0, 1, 1, 0],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 0, 1, 1],
            "post_sequence": [1, 0, 1, 0, 1, 0, 0, 1],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [0, 1, 0, 1, 0, 1, 0, 1],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 0, 1, 0, 1, 0, 1, 0],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 0, 1, 0, 1, 0, 1, 0],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 1, 0, 1, 1, 0, 1, 0],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 1, 0, 1, 1, 1, 0, 0],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [0, 1, 0, 1, 1, 0, 1, 1],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 0, 1],
            "post_sequence": [0, 1, 1, 0, 1, 0, 0, 1],
        },
        {
            "success": True,
            "pre_sequence": [1, 1, 1, 1, 1, 1, 1, 1],
            "post_sequence": [1, 0, 1, 0, 1, 0, 1, 0],
        },
    ],
    **REDUNDANT_JOB_DATA,
}


class MockClient:
    """Mock client for testing."""

    DEVICE_MAP = {
        "qbraid_qir_simulator": DEVICE_DATA_QIR,
        "quera_qasm_simulator": DEVICE_DATA_QUERA_QASM,
        "nec_vector_annealer": DEVICE_DATA_NEC,
        "quera_aquila": DEVICE_DATA_AQUILA,
    }

    JOB_MAP = {
        "qbraid_qir_simulator": JOB_DATA_QIR,
        "quera_qasm_simulator": JOB_DATA_QUERA_QASM,
        "nec_vector_annealer": JOB_DATA_NEC,
        "quera_aquila": JOB_DATA_AQUILA,
    }

    RESULTS_MAP = {
        "qbraid_qir_simulator": RESULTS_DATA_QIR,
        "quera_qasm_simulator": RESULTS_DATA_QUERA_QASM,
        "nec_vector_annealer": RESULTS_DATA_NEC,
        "quera_aquila": RESULTS_DATA_AQUILA,
    }

    @property
    def session(self):
        """Mock session property."""
        session = MagicMock()
        session.api_key = "abc123"
        return session

    @property
    def _user_metadata(self):
        """Mock user metadata property."""
        return {"organization": "qbraid", "role": "guest"}

    def search_devices(self, query: Optional[dict[str, Any]] = None) -> list[dict[str, Any]]:
        """Returns a list of devices matching the given query."""
        all_devices = [data.copy() for data in self.DEVICE_MAP.values()]

        if not query:
            return all_devices

        if "status" in query:
            valid_status_values = {status.name for status in DeviceStatus}
            if query["status"] not in valid_status_values:
                raise QuantumServiceRequestError("No devices found matching given criteria")

        devices = [
            device
            for device in all_devices
            if all(device.get(key) == value for key, value in query.items())
        ]

        return devices

    def _get_data(
        self, identifier: str, data_map: dict[str, dict[str, Any]], error_message: str
    ) -> dict[str, Any]:
        """Helper to retrieve data from a mapping, raising an error if not found."""
        try:
            return data_map[identifier].copy()
        except KeyError as err:
            raise QuantumServiceRequestError(error_message) from err

    def get_device(self, qbraid_id: str) -> dict[str, Any]:
        """Returns the device with the given ID."""
        return self._get_data(
            identifier=qbraid_id,
            data_map=self.DEVICE_MAP,
            error_message="No devices found matching given criteria",
        )

    def create_job(self, data: dict[str, Any]) -> dict[str, Any]:
        """Creates a new quantum job with the given data."""
        return self._get_data(
            identifier=data.get("qbraidDeviceId"),
            data_map=self.JOB_MAP,
            error_message="Failed to create job",
        )

    @staticmethod
    def _extract_device_id(job_id: str) -> str:
        """Extracts the device ID from the given job ID."""
        try:
            return job_id.split("-")[0]
        except IndexError as err:
            raise ValueError("Invalid job ID format") from err

    def get_job(self, job_id: str) -> dict[str, Any]:
        """Returns the quantum job with the given ID."""
        device_id = self._extract_device_id(job_id)
        return self._get_data(
            identifier=device_id,
            data_map=self.JOB_MAP,
            error_message="No jobs found matching given criteria",
        )

    # pylint: disable-next=unused-argument
    def get_job_results(self, job_id: str, **kwargs) -> dict[str, Any]:
        """Returns the results of the quantum job with the given ID."""
        device_id = self._extract_device_id(job_id)
        return self._get_data(
            identifier=device_id,
            data_map=self.RESULTS_MAP,
            error_message="Failed to retrieve job results",
        )


class MockDevice(QuantumDevice):
    """Mock basic device for testing."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def status(self):
        return DeviceStatus.ONLINE

    def submit(self, *args, **kwargs):
        raise NotImplementedError



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for defining custom Amazon Braket control gate

"""
import itertools
from typing import Any

import braket.ir.jaqcd as ir
import numpy as np
from braket.circuits import Gate, Instruction, QubitSet, circuit
from braket.circuits.gates import Unitary, format_complex
from braket.circuits.serialization import OpenQASMSerializationProperties


class C(Gate):
    """Controlled gate
    Args:
        sub_gate (Gate): Quantum Gate.
        targets (QubitSet): Target qubits.
    """

    def __init__(self, sub_gate: Gate, targets: QubitSet):
        self.sub_gate = sub_gate
        qubit_count = len(targets)
        sub_qubit_count = sub_gate.qubit_count
        self._num_controls = qubit_count - sub_qubit_count
        self._controls = targets[: self._num_controls]
        ascii_symbols = ["C"] * self._num_controls + list(self.sub_gate.ascii_symbols)

        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

    def _extend_matrix(self, sub_matrix: np.ndarray) -> np.ndarray:
        qid_shape = (2,) * self.qubit_count
        control_values = ((1,),) * self._num_controls
        sub_n = len(qid_shape) - self._num_controls
        tensor = np.eye(np.prod(qid_shape, dtype=np.int64).item(), dtype=complex)
        tensor.shape = qid_shape * 2
        sub_tensor = sub_matrix.reshape(qid_shape[self._num_controls :] * 2)
        for control_vals in itertools.product(*control_values):
            active = (*(v for v in control_vals), *(slice(None),) * sub_n) * 2
            tensor[active] = sub_tensor
        return tensor.reshape((np.prod(qid_shape, dtype=np.int64).item(),) * 2)

    def to_matrix(self, *args, **kwargs) -> np.ndarray:  # pylint: disable=unused-argument
        """Returns a matrix representation of the quantum operator

        Returns:
            np.ndarray: A matrix representation of the quantum operator
        """
        sub_matrix = self.sub_gate.to_matrix()
        return self._extend_matrix(sub_matrix)

    def adjoint(self) -> list[Gate]:
        """Returns the adjoint of the gate."""
        return [Unitary(self.to_matrix().conj().T, display_name=f"({self.ascii_symbols})^†")]

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.Unitary.construct(
            targets=list(target),
            matrix=Unitary._transform_matrix_to_ir(self.to_matrix()),
        )

    def _to_openqasm(  # pylint: disable=unused-argument
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties, **kwargs
    ) -> str:
        qubits = [serialization_properties.format_target(int(qubit)) for qubit in target]
        formatted_matrix = np.array2string(
            self.to_matrix(),
            separator=", ",
            formatter={"all": lambda x: format_complex(x)},  # pylint: disable=unnecessary-lambda
            threshold=float("inf"),
        ).replace("\n", "")

        return f"#pragma braket unitary({formatted_matrix}) {', '.join(qubits)}"

    @staticmethod
    @circuit.subroutine(register=True)
    def c(targets: QubitSet, sub_gate: Gate) -> Instruction:
        """Registers this function into the circuit class.
        Args:
            targets (QubitSet): Target qubits.
            sub_gate (Gate): Quantum Gate.
        Returns:
            Instruction: Controlled Gate Instruction.
        """
        return Instruction(C(sub_gate, targets), target=targets)


Gate.register_gate(C)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Pennylane Tapes used for testing

"""
import numpy as np
import pennylane as qml
from pennylane.tape import QuantumTape


def pennylane_bell() -> QuantumTape:
    """Returns Pennylane tape representing bell circuit"""
    with QuantumTape() as tape:
        qml.Hadamard(wires=0)
        qml.CNOT(wires=[0, 1])

    return tape


def pennylane_shared15() -> QuantumTape:
    """Returns Pennylane tape for shared gates tests."""
    with QuantumTape() as tape:
        qml.Hadamard(wires=0)
        qml.Hadamard(wires=1)
        qml.Hadamard(wires=2)
        qml.Hadamard(wires=3)
        qml.PauliX(wires=0)
        qml.PauliX(wires=1)
        qml.PauliY(wires=2)
        qml.PauliZ(wires=3)
        qml.S(wires=0)
        qml.adjoint(qml.S(wires=1))
        qml.T(wires=2)
        qml.adjoint(qml.T(wires=3))
        qml.RX(np.pi / 4, wires=0)
        qml.RY(np.pi / 2, wires=1)
        qml.RZ(3 * np.pi / 4, wires=2)
        qml.PhaseShift(np.pi / 8, wires=3)
        qml.SX(wires=0)
        qml.adjoint(qml.SX(wires=1))
        qml.ISWAP(wires=[2, 3])
        qml.SWAP(wires=[0, 2])
        qml.SWAP(wires=[1, 3])
        qml.CNOT(wires=[0, 1])
        qml.ControlledPhaseShift(np.pi / 4, wires=[2, 3])
    return tape



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for visualizing quantum programs, experimental results,
and other associated data.

.. currentmodule:: qbraid.visualization

Functions
----------

.. autosummary::
   :toctree: ../stubs/

    plot_histogram
    plot_distribution
    plot_conversion_graph
    plot_runtime_conversion_scheme
    circuit_drawer
    plot_atomic_register
    animate_qpu_state

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

    VisualizationError

"""
from .ahs import plot_atomic_register
from .draw_circuit import circuit_drawer
from .exceptions import VisualizationError
from .flair_animations import animate_qpu_state
from .plot_conversions import plot_conversion_graph, plot_runtime_conversion_scheme
from .plot_counts import plot_distribution, plot_histogram

__all__ = [
    "plot_histogram",
    "plot_distribution",
    "plot_conversion_graph",
    "plot_runtime_conversion_scheme",
    "circuit_drawer",
    "plot_atomic_register",
    "animate_qpu_state",
    "VisualizationError",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=too-many-arguments

"""
Plot atomic register of AHS program.

"""
from typing import Optional


def plot_atomic_register(
    sites: list[tuple[float, float]],
    filling: list[bool],
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    save_path: Optional[str] = None,
    show: bool = True,
) -> None:
    """
    Plots atomic register given site coordinates and filling.

    Args:
        sites (list[tuple[float, float]]): A list of tuples represents (x, y) coordinates.
        filling (list[bool]): A list of booleans indicating the filling status at each site.
        title (Optional[str]): The title of the plot. Defaults to "Atomic Register".
        xlabel (Optional[str]): The label for the x-axis. Defaults to "X (meters)".
        ylabel (Optional[str]): The label for the y-axis. Defaults to "Y (meters)".
        save_path (Optional[str]): Path to save plot. Plot is not saved unless specified.
        show (bool): If True, display the figure. Defaults to True.

    """
    # pylint: disable-next=import-outside-toplevel
    import matplotlib.pyplot as plt

    if not len(sites) == len(filling):
        raise ValueError("sites and filling must be of equal length.")

    xData = [x[0] for x in sites]
    yData = [x[1] for x in sites]

    plt.figure(figsize=(10, 10))
    facecolors = ["purple" if fill else "none" for fill in filling]
    plt.scatter(xData, yData, edgecolors="pink", facecolors=facecolors, s=100, zorder=5)

    plt.scatter([], [], edgecolor="pink", label="Atom", s=100, facecolors="purple")
    plt.scatter([], [], edgecolor="pink", label="No Atom", s=100, facecolors="none")

    plt.legend(loc="upper left")

    if title:
        plt.title(title, fontsize=16)
    if xlabel:
        plt.xlabel(xlabel, fontsize=14)
    if ylabel:
        plt.ylabel(ylabel, fontsize=14)

    plt.grid(False)
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path)

    if show:
        plt.show()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Pennylane OpenQASM conversions

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from pennylane.tape import QuantumTape

from qbraid.transpiler.annotations import weight

if TYPE_CHECKING:
    from qbraid.programs.typer import Qasm2StringType


@weight(1)
def pennylane_to_qasm2(tape: QuantumTape) -> Qasm2StringType:
    """Converts a PennyLane tape to OpenQASM 2.0

    Args:
        tape (QuantumTape): input PennyLane tape

    Returns:
        str: OpenQASM 2.0 representation of the tape

    """
    return tape.to_openqasm(measure_all=False)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for Azure Quantum runtime (remote) for QIR

"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from azure.quantum.target.microsoft import MicrosoftEstimatorResult
from qiskit import QuantumCircuit

from qbraid.runtime import DeviceStatus, JobStatus
from qbraid.runtime.azure import AzureQuantumProvider
from qbraid.transpiler.conversions.qiskit import qiskit_to_pyqir

if TYPE_CHECKING:
    import pyqir as pyqir_

pyqir = pytest.importorskip("pyqir")


@pytest.fixture
def qir_bitcode(qiskit_circuit: QuantumCircuit) -> bytes:
    """Fixture for QIR bitcode from a Qiskit quantum circuit."""
    module: pyqir_.Module = qiskit_to_pyqir(qiskit_circuit)
    return module.bitcode


@pytest.mark.skip(
    reason="The Resource Estimator is now fully open source and available as part of the QDK."
)
@pytest.mark.remote
@pytest.mark.parametrize("direct", [(True), (False)])
def test_submit_qir_to_microsoft(
    provider: AzureQuantumProvider, qiskit_circuit: QuantumCircuit, qir_bitcode: bytes, direct: bool
):
    """Test submitting Qiskit circuit or QIR bitcode to run on the Microsoft resource estimator."""
    device = provider.get_device("microsoft.estimator")
    assert device.status() == DeviceStatus.ONLINE

    input_params = {"entryPoint": qiskit_circuit.name, "arguments": [], "count": 100}

    if direct:
        job = device.submit(qir_bitcode, input_params=input_params)
    else:
        job = device.run(qiskit_circuit, input_params=input_params)

    job.wait_for_final_state()

    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, MicrosoftEstimatorResult)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Tests for qBraid transpiler conversion extras.

"""
import importlib.util
from typing import Callable

import braket.circuits
import pytest

try:
    import pyqir

    pyqir_installed = True
except ImportError:
    pyqir_installed = False


from qbraid.passes.qasm.compat import normalize_qasm_gate_params
from qbraid.transpiler.conversions.qasm3 import autoqasm_to_qasm3
from qbraid.transpiler.conversions.qiskit import qiskit_to_braket, qiskit_to_pyqir
from qbraid.transpiler.converter import transpile
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.graph import ConversionGraph


def has_extra(conversion_func: Callable) -> bool:
    """
    Check if the conversion function requires extra packages.

    Args:
        conversion_func (Callable): The conversion function to check for extra requirements.

    Returns:
        bool: True if all required extra packages are importable, False otherwise.
    """
    extras = getattr(conversion_func, "requires_extras", [])
    return all(importlib.util.find_spec(module_name) is not None for module_name in extras)


@pytest.mark.skipif(not has_extra(qiskit_to_braket), reason="Extra not installed")
@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_qiskit_to_braket_extra(bell_circuit):
    """Test qiskit-braket-provider transpiler conversion extra."""
    qiskit_circuit, _ = bell_circuit
    conversions = [Conversion("qiskit", "braket", qiskit_to_braket)]
    graph = ConversionGraph(conversions)
    program = transpile(qiskit_circuit, "braket", conversion_graph=graph, max_path_depth=1)
    assert isinstance(program, braket.circuits.Circuit)


@pytest.mark.skipif(not has_extra(qiskit_to_pyqir), reason="Extra not installed")
@pytest.mark.skipif(not pyqir_installed, reason="pyqir not installed")
@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_qiskit_to_pyqir_extra(bell_circuit):
    """Test qiskit-qir transpiler conversion extra."""
    qiskit_circuit, _ = bell_circuit
    conversions = [Conversion("qiskit", "pyqir", qiskit_to_pyqir)]
    graph = ConversionGraph(conversions)
    program = transpile(qiskit_circuit, "pyqir", conversion_graph=graph, max_path_depth=1)
    assert isinstance(program, pyqir.Module)


def autoqasm_bell_circuit():
    """Function that returns autoqasm bell circuit."""
    # pylint: disable-next=import-outside-toplevel
    from ..fixtures.autoqasm.circuits import autoqasm_bell

    return autoqasm_bell()


def qasm3_bell_reference():
    """Reference QASM3 string for Bell circuit"""
    return """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] __qubits__;
h __qubits__[0];
cx __qubits__[0], __qubits__[1];"""


@pytest.mark.skipif(not has_extra(autoqasm_to_qasm3), reason="Extra not installed")
def test_autoqasm_bell_to_qasm3_extra():
    """Test autoqasm-qasm3 conversion extra."""
    autoqasm_circuit = autoqasm_bell_circuit()
    conversions = [Conversion("autoqasm", "qasm3", autoqasm_to_qasm3)]
    graph = ConversionGraph(conversions)
    program = transpile(autoqasm_circuit, "qasm3", conversion_graph=graph, max_path_depth=1)
    assert isinstance(program, str)
    assert program == qasm3_bell_reference()


def autoqasm_shared15_circuit():
    """Function that returns autoqasm shared15 circuit."""
    # pylint: disable-next=import-outside-toplevel
    from ..fixtures.autoqasm.circuits import autoqasm_shared15

    return autoqasm_shared15()


def qasm3_shared15_reference():
    """Reference QASM3 string for shared15 circuit"""
    return """OPENQASM 3.0;
include "stdgates.inc";
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}
gate iswap _gate_q_0, _gate_q_1 {
  s _gate_q_0;
  s _gate_q_1;
  h _gate_q_0;
  cx _gate_q_0, _gate_q_1;
  cx _gate_q_1, _gate_q_0;
  h _gate_q_1;
}
qubit[4] __qubits__;
h __qubits__[0];
h __qubits__[1];
h __qubits__[2];
h __qubits__[3];
x __qubits__[0];
x __qubits__[1];
y __qubits__[2];
z __qubits__[3];
s __qubits__[0];
sdg __qubits__[1];
t __qubits__[2];
tdg __qubits__[3];
rx(pi/4) __qubits__[0];
ry(pi/2) __qubits__[1];
rz(3*pi/4) __qubits__[2];
p(pi/8) __qubits__[3];
sx __qubits__[0];
sxdg __qubits__[1];
iswap __qubits__[2], __qubits__[3];
swap __qubits__[0], __qubits__[2];
swap __qubits__[1], __qubits__[3];
cx __qubits__[0], __qubits__[1];
cp(pi/4) __qubits__[2], __qubits__[3];"""


@pytest.mark.skipif(not has_extra(autoqasm_to_qasm3), reason="Extra not installed")
def test_autoqasm_shared15_to_qasm3_extra():
    """Test autoqasm-qasm3 conversion extra."""
    autoqasm_circuit = autoqasm_shared15_circuit()
    conversions = [Conversion("autoqasm", "qasm3", autoqasm_to_qasm3)]
    graph = ConversionGraph(conversions)
    program = transpile(autoqasm_circuit, "qasm3", conversion_graph=graph, max_path_depth=1)

    assert isinstance(program, str)
    assert program == normalize_qasm_gate_params(qasm3_shared15_reference())



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing functions to convert from pyQuil's circuit
representation (Quil programs) to Cirq's circuit representation.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

from .cirq_quil_input import circuit_from_quil

if TYPE_CHECKING:
    import cirq.circuits
    import pyquil.quil


@weight(1)
def pyquil_to_cirq(program: pyquil.quil.Program) -> cirq.circuits.Circuit:
    """Returns a Cirq circuit equivalent to the input pyQuil Program.

    Args:
        program: PyQuil Program to convert to a Cirq circuit.

    Returns:
        Cirq circuit representation equivalent to the input pyQuil Program.
    """
    try:
        return circuit_from_quil(program.out())
    except Exception as err:
        raise ProgramConversionError(
            "qBraid transpiler doesn't yet support pyQuil noise gates."
        ) from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Qiskit OpenQASM conversions

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import weight

qiskit = LazyLoader("qiskit", globals(), "qiskit")

if TYPE_CHECKING:
    import qiskit as qiskit_

    from qbraid.programs.typer import Qasm2StringType


@weight(1)
def qasm2_to_qiskit(qasm: Qasm2StringType) -> qiskit_.QuantumCircuit:
    """Returns a Qiskit circuit equivalent to the input OpenQASM 2 string.

    Args:
        qasm: OpenQASM 2 string to convert to a Qiskit circuit.

    Returns:
        Qiskit.QuantumCircuit object equivalent to the input OpenQASM 2 string.
    """
    return qiskit.QuantumCircuit.from_qasm_str(qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing top-level qbraid program loader functionality
utilizing entrypoints via ``importlib``.

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Union

import openqasm3

from qbraid._entrypoints import load_entrypoint
from qbraid.exceptions import QbraidError

from .alias_manager import get_program_type_alias
from .exceptions import ProgramLoaderError
from .registry import QPROGRAM

if TYPE_CHECKING:
    from qbraid.programs.ahs import AnalogHamiltonianProgram
    from qbraid.programs.annealing import AnnealingProgram
    from qbraid.programs.gate_model import GateModelProgram


def load_program(
    program: QPROGRAM,
) -> Union[GateModelProgram, AnalogHamiltonianProgram, AnnealingProgram]:
    """Apply qbraid quantum program wrapper to a supported quantum program.

    This function is used to create a qBraid :class:`~qbraid.programs.QuantumProgram`
    object, which can then be transpiled to any supported quantum circuit-building package.
    The input quantum circuit object must be an instance of a circuit object derived from a
    supported package.

    Args:
        program: A supported quantum circuit/program object.

    Returns:
        QuantumProgram: A wrapped quantum program object of the inferred subclass.

    Raises:
        :class:`~qbraid.ProgramLoaderError`: If the input circuit is not a supported quantum program

    """
    if isinstance(program, openqasm3.ast.Program):
        program = openqasm3.dumps(program)

    try:
        package = get_program_type_alias(program)
    except QbraidError as err:
        raise ProgramLoaderError(f"Error loading quantum program of type {type(program)}") from err

    try:
        load_program_class = load_entrypoint("programs", package)
    except Exception as err:
        raise ProgramLoaderError(f"Error loading quantum program of type {type(program)}") from err

    program_instance = load_program_class(program)

    return program_instance



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for result builder utility functions for partial measurements

"""
from qbraid.runtime.aws.result_builder import marginal_count, marginal_measurement


def test_marginal_measurement_basic():
    """Test basic marginal measurement extraction."""
    measurements = [[0, 1, 0, 1], [1, 0, 1, 0], [0, 0, 1, 1]]
    qubit_indices = [0, 2]

    result = marginal_measurement(measurements, qubit_indices)
    expected = [[0, 0], [1, 1], [0, 1]]

    assert result == expected


def test_marginal_measurement_single_qubit():
    """Test marginal measurement with single qubit."""
    measurements = [[0, 1, 0], [1, 0, 1], [0, 1, 1]]
    qubit_indices = [1]

    result = marginal_measurement(measurements, qubit_indices)
    expected = [[1], [0], [1]]

    assert result == expected


def test_marginal_measurement_all_qubits():
    """Test marginal measurement with all qubits (should return original)."""
    measurements = [[0, 1, 0], [1, 0, 1]]
    qubit_indices = [0, 1, 2]

    result = marginal_measurement(measurements, qubit_indices)
    expected = [[0, 1, 0], [1, 0, 1]]

    assert result == expected


def test_marginal_measurement_reverse_order():
    """Test marginal measurement with indices in reverse order."""
    measurements = [[0, 1, 0, 1], [1, 0, 1, 0]]
    qubit_indices = [3, 1, 0]

    result = marginal_measurement(measurements, qubit_indices)
    expected = [[1, 1, 0], [0, 0, 1]]

    assert result == expected


def test_marginal_measurement_empty_measurements():
    """Test marginal measurement with empty measurements list."""
    measurements = []
    qubit_indices = [0, 1]

    result = marginal_measurement(measurements, qubit_indices)
    expected = []

    assert result == expected


def test_marginal_measurement_empty_indices():
    """Test marginal measurement with empty qubit indices."""
    measurements = [[0, 1, 0], [1, 0, 1]]
    qubit_indices = []

    result = marginal_measurement(measurements, qubit_indices)
    expected = [[], []]

    assert result == expected


def test_marginal_count_basic():
    """Test basic marginal count computation."""
    count_dict = {"0101": 10, "1010": 5, "0110": 3}
    qubit_indices = [0, 2]

    result = marginal_count(count_dict, qubit_indices)
    expected = {"00": 10, "11": 5, "01": 3}

    assert result == expected


def test_marginal_count_with_aggregation():
    """Test marginal count with aggregation of similar patterns."""
    count_dict = {"0101": 10, "0111": 5, "1101": 3, "1111": 2}
    qubit_indices = [0, 2]

    result = marginal_count(count_dict, qubit_indices)
    expected = {"00": 10, "01": 5, "10": 3, "11": 2}

    assert result == expected


def test_marginal_count_aggregation_same_pattern():
    """Test marginal count where different bitstrings map to same pattern."""
    count_dict = {"0101": 10, "0001": 5, "0111": 3, "0011": 2}
    qubit_indices = [0, 3]  # First and last bits

    result = marginal_count(count_dict, qubit_indices)
    # "0101" -> "01", "0001" -> "01", "0111" -> "01", "0011" -> "01"
    expected = {"01": 20}  # 10 + 5 + 3 + 2

    assert result == expected


def test_marginal_count_single_qubit():
    """Test marginal count with single qubit."""
    count_dict = {"010": 10, "110": 5, "001": 3}
    qubit_indices = [1]

    result = marginal_count(count_dict, qubit_indices)
    expected = {"1": 15, "0": 3}  # "010"->1, "110"->1, "001"->0

    assert result == expected


def test_marginal_count_all_qubits():
    """Test marginal count with all qubits (should return original)."""
    count_dict = {"010": 10, "110": 5, "001": 3}
    qubit_indices = [0, 1, 2]

    result = marginal_count(count_dict, qubit_indices)
    expected = {"010": 10, "110": 5, "001": 3}

    assert result == expected


def test_marginal_count_empty_dict():
    """Test marginal count with empty count dictionary."""
    count_dict = {}
    qubit_indices = [0, 1]

    result = marginal_count(count_dict, qubit_indices)
    expected = {}

    assert result == expected


def test_marginal_count_empty_indices():
    """Test marginal count with empty qubit indices."""
    count_dict = {"010": 10, "110": 5}
    qubit_indices = []

    result = marginal_count(count_dict, qubit_indices)
    expected = {"": 15}  # All counts aggregated to empty string

    assert result == expected


def test_marginal_count_complex_aggregation():
    """Test complex marginal count with multiple aggregations."""
    count_dict = {
        "0000": 10,
        "0001": 5,
        "0010": 3,
        "0011": 2,
        "1000": 8,
        "1001": 4,
        "1010": 6,
        "1011": 1,
    }
    qubit_indices = [0, 2]  # First and third bits

    result = marginal_count(count_dict, qubit_indices)
    # Let me recalculate:
    # "0000" -> bits [0,2] -> "00" (10)
    # "0001" -> bits [0,2] -> "00" (5)
    # "0010" -> bits [0,2] -> "01" (3)
    # "0011" -> bits [0,2] -> "01" (2)
    # "1000" -> bits [0,2] -> "10" (8)
    # "1001" -> bits [0,2] -> "10" (4)
    # "1010" -> bits [0,2] -> "11" (6)
    # "1011" -> bits [0,2] -> "11" (1)
    expected = {"00": 15, "01": 5, "10": 12, "11": 7}  # 10 + 5  # 3 + 2  # 8 + 4  # 6 + 1

    assert result == expected


def test_marginal_count_reverse_order_indices():
    """Test marginal count with indices in reverse order."""
    count_dict = {"0101": 10, "1010": 5}
    qubit_indices = [3, 1, 0]  # Reverse order

    result = marginal_count(count_dict, qubit_indices)
    expected = {"110": 10, "001": 5}  # "0101"->bits[3,1,0]->"110", "1010"->bits[3,1,0]->"001"

    assert result == expected



# Copyright (C) 2024 qBraid
# Copyright (C) 2022 The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for testing qBraid QuilOutput.

"""
import os

import cirq
import numpy as np
import pytest
from cirq.ops.pauli_interaction_gate import PauliInteractionGate

try:
    from qbraid.transpiler.conversions.cirq.cirq_quil_output import (
        QuilFormatter,
        QuilOneQubitGate,
        QuilOutput,
        QuilTwoQubitGate,
        _quil_u2_gate,
        _quil_u3_gate,
        _rzz_gate,
        _twoqubitdiagonal_gate,
        exponent_to_pi_string,
    )
    from qbraid.transpiler.conversions.qasm2.cirq_custom import RZZGate, U2Gate, U3Gate

    pyquil_not_installed = False
except ImportError:
    pyquil_not_installed = True

pytestmark = pytest.mark.skipif(pyquil_not_installed, reason="pyquil not installed")


def _make_qubits(n):
    """Create a list of named qubits."""
    return [cirq.NamedQubit(f"q{i}") for i in range(n)]


def test_single_gate_no_parameter():
    """Test that a single gate with no parameter is correctly converted to Quil."""
    (q0,) = _make_qubits(1)
    output = QuilOutput((cirq.X(q0),), (q0,))
    assert (
        str(output)
        == """# Created using qBraid.

X 0\n"""
    )


def test_single_gate_with_parameter():
    """Test that a single gate with a parameter is correctly converted to Quil."""
    (q0,) = _make_qubits(1)
    output = QuilOutput((cirq.X(q0) ** 0.5,), (q0,))
    assert (
        str(output)
        == f"""# Created using qBraid.

RX({np.pi / 2}) 0\n"""
    )


def test_single_gate_named_qubit():
    """Test that a single gate with a named qubit is correctly converted to Quil."""
    q = cirq.NamedQubit("qTest")
    output = QuilOutput((cirq.X(q),), (q,))

    assert (
        str(output)
        == """# Created using qBraid.

X 0\n"""
    )


def test_h_gate_with_parameter():
    """Test that a Hadamard gate with a parameter is correctly converted to Quil."""
    (q0,) = _make_qubits(1)
    output = QuilOutput((cirq.H(q0) ** 0.25,), (q0,))
    assert (
        str(output)
        == f"""# Created using qBraid.

RY({np.pi / 4}) 0
RX({np.pi / 4}) 0
RY({-np.pi / 4}) 0\n"""
    )


def test_save_to_file(tmpdir):
    """Test that a QuilOutput can be saved to a file."""
    file_path = os.path.join(tmpdir, "test.quil")
    (q0,) = _make_qubits(1)
    output = QuilOutput((cirq.X(q0)), (q0,))
    output.save_to_file(file_path)
    with open(file_path, "r", encoding="utf-8") as f:
        file_content = f.read()
    assert (
        file_content
        == """# Created using qBraid.

X 0\n"""
    )


def test_quil_one_qubit_gate_repr():
    """Test that the QuilOneQubitGate __repr__ method works as expected."""
    gate = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    assert repr(gate) == (
        """cirq.circuits.quil_output.QuilOneQubitGate(matrix=
[[1 0]
 [0 1]]
)"""
    )


def test_quil_two_qubit_gate_repr():
    """Test that the QuilTwoQubitGate __repr__ method works as expected."""
    gate = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))
    assert repr(gate) == (
        """cirq.circuits.quil_output.QuilTwoQubitGate(matrix=
[[1 0 0 0]
 [0 1 0 0]
 [0 0 1 0]
 [0 0 0 1]]
)"""
    )


def test_quil_one_qubit_gate_eq():
    """Test that the QuilOneQubitGate __eq__ method works as expected."""
    gate = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    gate2 = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    assert cirq.approx_eq(gate, gate2, atol=1e-16)
    gate3 = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    gate4 = QuilOneQubitGate(np.array([[1, 0], [0, 2]]))
    assert not cirq.approx_eq(gate4, gate3, atol=1e-16)


def test_quil_two_qubit_gate_eq():
    """Test that the QuilTwoQubitGate __eq__ method works as expected."""
    gate = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))
    gate2 = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))
    assert cirq.approx_eq(gate, gate2, atol=1e-8)
    gate3 = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))
    gate4 = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 2, 0], [0, 0, 0, 1]]))
    assert not cirq.approx_eq(gate4, gate3, atol=1e-8)


def test_quil_one_qubit_gate_output():
    """Test that a QuilOneQubitGate can be correctly converted to Quil."""
    (q0,) = _make_qubits(1)
    gate = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    output = QuilOutput((gate.on(q0),), (q0,))
    assert (
        str(output)
        == """# Created using qBraid.

DEFGATE USERGATE1:
    1.0+0.0i, 0.0+0.0i
    0.0+0.0i, 1.0+0.0i
USERGATE1 0
"""
    )


def test_two_quil_one_qubit_gate_output():
    """Test that two QuilOneQubitGates can be correctly converted to Quil."""
    (q0,) = _make_qubits(1)
    gate = QuilOneQubitGate(np.array([[1, 0], [0, 1]]))
    gate1 = QuilOneQubitGate(np.array([[2, 0], [0, 3]]))
    output = QuilOutput((gate.on(q0), gate1.on(q0)), (q0,))
    assert (
        str(output)
        == """# Created using qBraid.

DEFGATE USERGATE1:
    1.0+0.0i, 0.0+0.0i
    0.0+0.0i, 1.0+0.0i
USERGATE1 0
DEFGATE USERGATE2:
    2.0+0.0i, 0.0+0.0i
    0.0+0.0i, 3.0+0.0i
USERGATE2 0
"""
    )


def test_quil_two_qubit_gate_output():
    """Test that a QuilTwoQubitGate can be correctly converted to Quil."""
    (q0, q1) = _make_qubits(2)
    gate = QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))
    output = QuilOutput((gate.on(q0, q1),), (q0, q1))
    assert (
        str(output)
        == """# Created using qBraid.

DEFGATE USERGATE1:
    1.0+0.0i, 0.0+0.0i, 0.0+0.0i, 0.0+0.0i
    0.0+0.0i, 1.0+0.0i, 0.0+0.0i, 0.0+0.0i
    0.0+0.0i, 0.0+0.0i, 1.0+0.0i, 0.0+0.0i
    0.0+0.0i, 0.0+0.0i, 0.0+0.0i, 1.0+0.0i
USERGATE1 0 1
"""
    )


def test_unsupported_operation():
    """Test that an unsupported operation raises an error."""
    (q0,) = _make_qubits(1)

    class UnsupportedOperation(cirq.Operation):
        """Mock class for an unsupported operation."""

        qubits = (q0,)
        with_qubits = NotImplemented

    output = QuilOutput((UnsupportedOperation(),), (q0,))
    with pytest.raises(ValueError):
        _ = str(output)


def test_i_swap_with_power():
    """Test that an ISWAP gate with a power is correctly converted to Quil."""
    q0, q1 = _make_qubits(2)

    output = QuilOutput((cirq.ISWAP(q0, q1) ** 0.25,), (q0, q1))
    assert (
        str(output)
        == f"""# Created using qBraid.

XY({np.pi / 4}) 0 1
"""
    )


# pylint: disable-next=unused-argument
def _all_operations(q0, q1, q2, q3, q4):
    """Create a list of all supported operations."""
    return (
        cirq.Z(q0),
        cirq.Z(q0) ** 0.625,
        cirq.Y(q0),
        cirq.Y(q0) ** 0.375,
        cirq.X(q0),
        cirq.X(q0) ** 0.875,
        cirq.H(q1),
        cirq.CZ(q0, q1),
        cirq.CZ(q0, q1) ** 0.25,  # Requires 2-qubit decomposition
        cirq.CNOT(q0, q1),
        cirq.CNOT(q0, q1) ** 0.5,  # Requires 2-qubit decomposition
        cirq.SWAP(q0, q1),
        cirq.SWAP(q1, q0) ** -1,
        cirq.SWAP(q0, q1) ** 0.75,  # Requires 2-qubit decomposition
        cirq.CCZ(q0, q1, q2),
        cirq.CCX(q0, q1, q2),
        cirq.CCZ(q0, q1, q2) ** 0.5,
        cirq.CCX(q0, q1, q2) ** 0.5,
        cirq.CSWAP(q0, q1, q2),
        cirq.XX(q0, q1),
        cirq.XX(q0, q1) ** 0.75,
        cirq.YY(q0, q1),
        cirq.YY(q0, q1) ** 0.75,
        cirq.ZZ(q0, q1),
        cirq.ZZ(q0, q1) ** 0.75,
        cirq.IdentityGate(1).on(q0),
        cirq.IdentityGate(3).on(q0, q1, q2),
        cirq.ISWAP(q2, q0),  # Requires 2-qubit decomposition
        cirq.PhasedXPowGate(phase_exponent=0.111, exponent=0.25).on(q1),
        cirq.PhasedXPowGate(phase_exponent=0.333, exponent=0.5).on(q1),
        cirq.PhasedXPowGate(phase_exponent=0.777, exponent=-0.5).on(q1),
        cirq.wait(q0, nanos=0),
        cirq.measure(q0, key="xX"),
        cirq.measure(q2, key="x_a"),
        cirq.measure(q3, key="X"),
        cirq.measure(q2, key="x_a"),
        cirq.measure(q1, q2, q3, key="multi", invert_mask=(False, True)),
    )


def test_all_operations():
    """Test that all operations are correctly converted to Quil."""
    qubits = tuple(_make_qubits(5))
    operations = _all_operations(*qubits)
    output = QuilOutput(operations, qubits)

    assert (
        str(output)
        == f"""# Created using qBraid.

DECLARE m0 BIT[1]
DECLARE m1 BIT[1]
DECLARE m2 BIT[1]
DECLARE m3 BIT[3]

Z 0
RZ({5 * np.pi / 8}) 0
Y 0
RY({3 * np.pi / 8}) 0
X 0
RX({7 * np.pi / 8}) 0
H 1
CZ 0 1
CPHASE({np.pi / 4}) 0 1
CNOT 0 1
RY({-np.pi / 2}) 1
CPHASE({np.pi / 2}) 0 1
RY({np.pi / 2}) 1
SWAP 0 1
SWAP 1 0
PSWAP({3 * np.pi / 4}) 0 1
H 2
CCNOT 0 1 2
H 2
CCNOT 0 1 2
RZ({np.pi / 8}) 0
RZ({np.pi / 8}) 1
RZ({np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 1
RZ({np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
H 2
RZ({np.pi / 8}) 0
RZ({np.pi / 8}) 1
RZ({np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 1
RZ({np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
RZ({-np.pi / 8}) 2
CNOT 0 1
CNOT 1 2
H 2
CSWAP 0 1 2
X 0
X 1
RX({3 * np.pi / 4}) 0
RX({3 * np.pi / 4}) 1
Y 0
Y 1
RY({3 * np.pi / 4}) 0
RY({3 * np.pi / 4}) 1
Z 0
Z 1
RZ({3 * np.pi / 4}) 0
RZ({3 * np.pi / 4}) 1
I 0
I 0
I 1
I 2
ISWAP 2 0
RZ({-0.111 * np.pi}) 1
RX({np.pi / 4}) 1
RZ({0.111 * np.pi}) 1
RZ({-0.333 * np.pi}) 1
RX({np.pi / 2}) 1
RZ({0.333 * np.pi}) 1
RZ({-0.777 * np.pi}) 1
RX({-np.pi / 2}) 1
RZ({0.777 * np.pi}) 1
WAIT
MEASURE 0 m0[0]
MEASURE 2 m1[0]
MEASURE 3 m2[0]
MEASURE 2 m1[0]
MEASURE 1 m3[0]
X 2 # Inverting for following measurement
MEASURE 2 m3[1]
MEASURE 3 m3[2]
"""
    )


def test_fails_on_big_unknowns():
    """Test that an error is raised when trying to convert an operation that is not supported."""

    class UnrecognizedGate(cirq.testing.ThreeQubitGate):
        """Mock class for an unsupported 3 qubit gate."""

    q0, q1, q2 = _make_qubits(3)
    res = QuilOutput(UnrecognizedGate().on(q0, q1, q2), (q0, q1, q2))
    with pytest.raises(ValueError, match="Cannot output operation as QUIL"):
        _ = str(res)


def test_pauli_interaction_gate():
    """Test that a PauliInteractionGate is correctly converted to Quil."""
    (q0, q1) = _make_qubits(2)
    output = QuilOutput(PauliInteractionGate.CZ.on(q0, q1), (q0, q1))
    assert (
        str(output)
        == """# Created using qBraid.

CZ 0 1
"""
    )


def test_equivalent_unitaries():
    """This test covers the factor of pi change. However, it will be skipped
    if pyquil is unavailable for import.

    References:
        https://docs.pytest.org/en/latest/skipping.html#skipping-on-a-missing-import-dependency
    """
    pyquil = pytest.importorskip("pyquil")
    pyquil_simulation_tools = pytest.importorskip("pyquil.simulation.tools")
    q0, q1 = _make_qubits(2)
    operations = [
        cirq.XPowGate(exponent=0.5, global_shift=-0.5)(q0),
        cirq.YPowGate(exponent=0.5, global_shift=-0.5)(q0),
        cirq.ZPowGate(exponent=0.5, global_shift=-0.5)(q0),
        cirq.CZPowGate(exponent=0.5)(q0, q1),
        cirq.ISwapPowGate(exponent=0.5)(q0, q1),
    ]
    output = QuilOutput(operations, (q0, q1))
    program = pyquil.Program(str(output))
    pyquil_unitary = pyquil_simulation_tools.program_unitary(program, n_qubits=2)
    # Qubit ordering differs between pyQuil and Cirq.
    cirq_unitary = cirq.Circuit(cirq.SWAP(q0, q1), operations, cirq.SWAP(q0, q1)).unitary()
    assert np.allclose(pyquil_unitary, cirq_unitary)


QUIL_CPHASES_PROGRAM = """
CPHASE00(pi/2) 0 1
CPHASE01(pi/2) 0 1
CPHASE10(pi/2) 0 1
CPHASE(pi/2) 0 1
"""

QUIL_DIAGONAL_DECOMPOSE_PROGRAM = """
RZ(0) 0
RZ(0) 1
CPHASE(0) 0 1
X 0
X 1
CPHASE(0) 0 1
X 0
X 1
"""


def test_two_qubit_diagonal_gate_quil_output():
    """Test that a TwoQubitDiagonalGate is correctly converted to Quil."""
    pyquil = pytest.importorskip("pyquil")
    pyquil_simulation_tools = pytest.importorskip("pyquil.simulation.tools")
    q0, q1 = _make_qubits(2)
    operations = [
        cirq.TwoQubitDiagonalGate([np.pi / 2, 0, 0, 0])(q0, q1),
        cirq.TwoQubitDiagonalGate([0, np.pi / 2, 0, 0])(q0, q1),
        cirq.TwoQubitDiagonalGate([0, 0, np.pi / 2, 0])(q0, q1),
        cirq.TwoQubitDiagonalGate([0, 0, 0, np.pi / 2])(q0, q1),
    ]
    output = QuilOutput(operations, (q0, q1))
    program = pyquil.Program(str(output))
    assert f"\n{program.out()}" == QUIL_CPHASES_PROGRAM

    pyquil_unitary = pyquil_simulation_tools.program_unitary(program, n_qubits=2)
    # Qubit ordering differs between pyQuil and Cirq.
    cirq_unitary = cirq.Circuit(cirq.SWAP(q0, q1), operations, cirq.SWAP(q0, q1)).unitary()
    assert np.allclose(pyquil_unitary, cirq_unitary)
    # Also test non-CPHASE case, which decomposes into X/RZ/CPhase
    operations = [cirq.TwoQubitDiagonalGate([0, 0, 0, 0])(q0, q1)]
    output = QuilOutput(operations, (q0, q1))
    program = pyquil.Program(str(output))
    assert f"\n{program.out()}" == QUIL_DIAGONAL_DECOMPOSE_PROGRAM


def test_parseable_defgate_output():
    """Test that a QuilOutput with a DEFGATE definition is correctly parsed."""
    pyquil = pytest.importorskip("pyquil")
    q0, q1 = _make_qubits(2)
    operations = [
        QuilOneQubitGate(np.array([[1, 0], [0, 1]])).on(q0),
        QuilTwoQubitGate(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])).on(
            q0, q1
        ),
    ]
    output = QuilOutput(operations, (q0, q1))
    # Just checks that we can create a pyQuil Program without crashing.
    pyquil.Program(str(output))


def test_unconveritble_op():
    """Test that an operation that cannot be converted to Quil raises an error."""
    (q0,) = _make_qubits(1)

    class MyGate(cirq.Gate):
        """Mock class for an one qubit gate."""

        def num_qubits(self) -> int:
            """Return the number of qubits."""
            return 1

    op = MyGate()(q0)

    # Documenting that this
    # operation would crash if you call _op_to_quil_directly
    with pytest.raises(ValueError, match="Can't convert"):
        _ = QuilOutput(op, (q0,))._op_to_quil(op)


@pytest.mark.parametrize(
    "input_exp, expected",
    [
        (0.25 * np.pi, "pi/4"),
        (-0.25 * np.pi, "-pi/4"),
        (0, "0"),
        (np.pi / 3, "pi/3"),
        (-1.25 * np.pi, "-5*pi/4"),
        (1 * np.pi, "pi"),
        (-1 * np.pi, "-pi"),
    ],
)
def test_exponent_to_pi_string(input_exp, expected):
    """Test that an exponent is correctly converted to a string with pi."""
    assert exponent_to_pi_string(input_exp) == expected


def test_two_qubit_diagonal_gate_none():
    """ "Test that _twoqubitdiagonal_gate returns None for non-diagonal gates."""
    (q0, q1) = _make_qubits(2)
    gate = cirq.TwoQubitDiagonalGate(np.array([0, 0, 0, 0]))
    formatter = QuilFormatter({q0: "q0", q1: "q1"}, {})
    circuit = cirq.Circuit()
    circuit.append(gate.on(q0, q1))
    instr = next(circuit.all_operations())
    assert _twoqubitdiagonal_gate(instr, formatter) is None


def test_rzz_rads_0_to_identity():
    """ "Test that RZZ gate with rads=0 maps to pyquil Identity gate."""
    (q0, q1) = _make_qubits(2)
    gate = RZZGate(rads=0)
    formatter = QuilFormatter({q0: "0", q1: "1"}, {})
    circuit = cirq.Circuit()
    circuit.append(gate.on(q0, q1))
    instr = next(circuit.all_operations())
    assert _rzz_gate(instr, formatter) == "I 0\nI 1\n"


@pytest.mark.parametrize(
    "gate_class, params, expected_quil, quil_converter",
    [
        (
            U2Gate,
            (np.pi / 4, np.pi / 8),
            "U(1.5707963267948966, 0.7853981633974483, 0.39269908169872414) 0",
            _quil_u2_gate,
        ),
        (
            U3Gate,
            (np.pi / 2, np.pi / 4, np.pi / 8),
            "U(1.5707963267948966, 0.7853981633974483, 0.39269908169872414) 0",
            _quil_u3_gate,
        ),
    ],
)
def test_custom_u_gate_to_quil(gate_class, params, expected_quil, quil_converter):
    """Test converting custom gates (U2, U3) to Quil."""
    (q0,) = _make_qubits(1)
    gate = gate_class(*params)
    formatter = QuilFormatter({q0: "0"}, {})
    circuit = cirq.Circuit()
    circuit.append(gate.on(q0))
    instr = next(circuit.all_operations())
    assert quil_converter(instr, formatter).strip() == expected_quil



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for plotting historgram of measurement counts against quantum states.

"""

from typing import Callable, Optional, Union

from qbraid_core._import import LazyLoader

plt = LazyLoader("plt", globals(), "matplotlib.pyplot")

# pylint: disable=too-many-arguments,unnecessary-lambda


def _counts_to_decimal(counts: dict) -> dict:
    """
    Converts a dictionary of counts to decimal form.

    Args:
        counts_dict (dict): A dictionary where keys are strings representing states and
                            values are integers representing counts.

    Returns:
        dict: A dictionary with the same keys as the input dictionary, but with values
              converted to their respective decimal proportions.

    Raises:
        ValueError: If the total count is zero.
        TypeError:  If the input dictionary contains non-integer values

    Example:
        >>> counts_to_decimal({"00": 10, "01": 15, "10": 20, "11": 5})
        {"00": 0.2, "01": 0.3, "10": 0.4, "11": 0.1}
    """

    try:
        total_count = sum(counts.values())
    except TypeError as err:
        raise TypeError("Counts values must be integers.") from err

    if total_count == 0:
        raise ValueError("Total count cannot be zero.")

    decimal_dict = {key: value / total_count for key, value in counts.items()}

    return decimal_dict


def _plot_data(
    counts: Union[list[dict], dict],
    legend: Optional[Union[list[str], str]] = None,
    colors: Optional[Union[list[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
    transform_fn: Optional[Callable[[dict], dict]] = None,
    label_format_fn: Optional[Callable[[float], str]] = lambda x: str(x),
):
    if not isinstance(counts, list):
        counts = [counts]
        if isinstance(colors, str):
            colors = [colors]
        if isinstance(legend, str):
            legend = [legend]

    if transform_fn:
        counts = [transform_fn(counts_dict) for counts_dict in counts]

    all_states = sorted(set(state for counts_dict in counts for state in counts_dict.keys()))

    num_dicts = len(counts)
    bar_width = 0.8 / num_dicts

    x_positions = range(len(all_states))

    if colors is None:
        cmap = plt.get_cmap("tab10")
        colors = [cmap(i / 10) for i in range(num_dicts)]

    if len(colors) != len(counts):
        raise ValueError("Number of colors must match number of datasets")

    if isinstance(legend, list) and len(legend) != len(counts):
        raise ValueError("Number of legend labels must match number of datasets")

    for i, counts_dict in enumerate(counts):
        counts_iter = [counts_dict.get(state, 0) for state in all_states]

        default_label = f"Job {i}" if num_dicts > 1 else None
        label = legend[i] if legend and i < len(legend) else default_label

        bars = plt.bar(
            [x + (i * bar_width) for x in x_positions],
            counts_iter,
            width=bar_width,
            color=colors[i % num_dicts],
            label=label,
            align="center",
        )

        y_min, y_max = plt.gca().get_ylim()
        y_range = y_max - y_min
        offset_percentage = 0.02  # Adjust this value to get the desired offset
        y_offset = y_range * offset_percentage

        for hbar, count in zip(bars, counts_iter):
            plt.text(
                hbar.get_x() + hbar.get_width() / 2,
                hbar.get_height() + +y_offset,
                label_format_fn(count),
                ha="center",
                va="bottom",
                color="black",
                fontsize=8,
            )

    plt.xticks(x_positions, all_states, rotation=45)

    y_ticks = plt.gca().get_yticks()
    plt.yticks(y_ticks)
    plt.grid(axis="y", linestyle="--", linewidth=0.7, alpha=0.7)

    if y_label:
        plt.ylabel(y_label)

    if x_label:
        plt.xlabel(x_label)

    if title:
        plt.title(title)

    if legend or num_dicts > 1:
        plt.legend()

    if save_path:
        plt.savefig(save_path)

    if show_plot:
        plt.show()


def plot_distribution(
    counts: Union[list[dict], dict],
    legend: Optional[Union[list[str], str]] = None,
    colors: Optional[Union[list[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
):
    """
    Plots a histogram probability distribution of quantum states.

    Args:
        counts (Union[list[Dict], Dict]): Dictionary or a list of dictionaries containing the
                                          quantum states as keys and their respective counts as
                                          values.
        legend (Optional[Union[list[str], str]]): List of strings or a single string representing
                                                  the labels of the datasets. Defaults to None,
                                                  where it generates default labels.
        colors (Optional[Union[list[str], str]]): List of strings or a single string representing
                                                  the colors for each dataset. Defaults to None,
                                                  where it generates a color sequence.
        title (Optional[str]): String representing the title of the plot. Defaults to None.
        x_label (Optional[str]): String representing the label for the x-axis. Defaults to None.
        y_label (Optional[str]): String representing the label for the y-axis. Defaults to None.
        show_plot (Optional[bool]): Boolean representing whether to show the plot. Defaults to True.
        save_path (Optional[str]): String representing the path to save the plot. Defaults to None.

    Returns:
        None: This function does not return a value; it displays a plot.

    Raises:
        ValueError: Raises an error if input arguments do not match the expected types or formats.

    Example:
        .. code-block:: python

            counts_dict1 = {'00': 50, '01': 30, '10': 10, '11': 10}
            counts_dict2 = {'00': 20, '01': 40, '10': 30, '11': 10}

            plot_distribution(
                counts=[counts_dict1, counts_dict2],
                legend=['First Execution', 'Second Execution'],
                colors=['crimson', 'midnightblue'],
                title="Quantum State Probability Distribution",
                x_label="Quantum States",
                y_label="Probabilities"
            )
    """

    y_label = "Quasi-probabilities" if not y_label else y_label

    _plot_data(
        counts,
        legend,
        colors,
        title,
        x_label,
        y_label,
        show_plot,
        save_path,
        transform_fn=_counts_to_decimal,
        label_format_fn=lambda x: "{:.3f}".format(x),  # pylint: disable=consider-using-f-string
    )


def plot_histogram(
    counts: Union[list[dict], dict],
    legend: Optional[Union[list[str], str]] = None,
    colors: Optional[Union[list[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
):
    """
    Plots a histogram of measurement counts against quantum states.

    Args:
        counts (Union[list[Dict], Dict]): Dictionary or a list of dictionaries containing the
                                          quantum states as keys and their respective counts as
                                          values.
        legend (Optional[Union[list[str], str]]): List of strings or a single string representing
                                                  the labels of the datasets. Defaults to None,
                                                  where it generates default labels.
        colors (Optional[Union[list[str], str]]): List of strings or a single string representing
                                                  the colors for each dataset. Defaults to None,
                                                  where it generates a color sequence.
        title (Optional[str]): String representing the title of the plot. Defaults to None.
        x_label (Optional[str]): String representing the label for the x-axis. Defaults to None.
        y_label (Optional[str]): String representing the label for the y-axis. Defaults to None.
        show_plot (Optional[bool]): Boolean representing whether to show the plot. Defaults to True.
        save_path (Optional[str]): String representing the path to save the plot. Defaults to None.

    Returns:
        None: This function does not return a value; it displays a plot.

    Raises:
        ValueError: Raises an error if input arguments do not match the expected types or formats.

    Example:
        .. code-block:: python

            counts_dict1 = {'00': 50, '01': 30, '10': 10, '11': 10}
            counts_dict2 = {'00': 20, '01': 40, '10': 30, '11': 10}

            plot_histogram(
                counts=[counts_dict1, counts_dict2],
                legend=['First Execution', 'Second Execution'],
                colors=['crimson', 'midnightblue'],
                title="Quantum State Measurement Counts",
                x_label="Quantum States",
                y_label="Counts"
            )
    """

    y_label = "Counts" if not y_label else y_label

    _plot_data(
        counts,
        legend,
        colors,
        title,
        x_label,
        y_label,
        show_plot,
        save_path,
        label_format_fn=lambda x: str(x),
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for qbraid.programs.qasm.OpenQasm2Program

"""
import textwrap
from unittest.mock import MagicMock

import openqasm3.ast
import pytest
from openqasm3.parser import parse

from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.qasm2 import OpenQasm2Program
from qbraid.programs.loader import load_program
from qbraid.programs.registry import unregister_program_type
from qbraid.programs.typer import Qasm2String

from ..fixtures.qasm2.circuits import qasm2_bell, qasm2_shared15


def test_qasm_qubits():
    """Test getting QASM qubits"""
    program1 = OpenQasm2Program(qasm2_bell())
    program1.validate()
    assert program1.qubits == {"q": 2}

    program2 = OpenQasm2Program(qasm2_shared15())
    program2.validate()
    assert program2.qubits == {"q": 4}


def test_qasm_num_qubits():
    """Test calculating number of qubits in qasm2 circuit"""
    assert OpenQasm2Program(qasm2_bell()).num_qubits == 2
    assert OpenQasm2Program(qasm2_shared15()).num_qubits == 4


def test_qasm_num_clbits():
    """Test calculating number of classical bits in qasm2 circuit"""
    assert OpenQasm2Program(qasm2_shared15()).num_clbits == 4


def test_qasm_depth():
    """Test calculating depth of qasm2 circuit"""
    assert OpenQasm2Program(qasm2_bell()).depth == 2
    assert OpenQasm2Program(qasm2_shared15()).depth == 12


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    try:
        with pytest.raises(ProgramTypeError):
            OpenQasm2Program({})
    finally:
        unregister_program_type("dict")


@pytest.fixture
def simple_qasm():
    """Fixture for a simple qasm2 circuit with measurements"""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[2];
    creg c[2];

    h q[0];

    measure q[0] -> c[0];
    measure q[1] -> c[1];
    """
    qasm = textwrap.dedent(qasm).strip()
    return qasm


def test_num_classical_bits(simple_qasm):
    """Test calculating number of classical bits in qasm2 circuit"""

    assert OpenQasm2Program(simple_qasm).num_clbits == 2


def test_remove_measurements_via_transform(simple_qasm):
    """Test removing measurements via transform method"""
    device = MagicMock()
    device.id = "quera_qasm_simulator"
    qprogram = OpenQasm2Program(simple_qasm)
    qprogram.transform(device=device)
    assert qprogram._module.has_measurements() is False
    assert isinstance(qprogram.program, Qasm2String)


def test_load_qasm2_program_from_parsed_obj(simple_qasm):
    """Test loading a qasm2 program from a openqasm3.ast.Program object"""
    parsed_program = parse(simple_qasm)
    assert isinstance(parsed_program, openqasm3.ast.Program)

    qbraid_program = load_program(parsed_program)
    assert isinstance(qbraid_program, OpenQasm2Program)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Display information about quantum jobs in table view.

"""

from typing import Optional

try:
    from IPython.display import HTML, clear_output, display
except ImportError:  # pragma: no cover
    pass

from qbraid._display import running_in_jupyter


def _job_table_basic(data: list[str], message: str) -> None:
    if len(data) == 0:
        print(message)
    else:
        headers = ["Job ID", "Submitted", "Status"]

        max_job_id_length = max(len(str(job_id)) for job_id, _, _ in data)
        max_job_id_length = max(max_job_id_length, len(headers[0]))

        header_format = "{:<" + str(max_job_id_length) + "} {:<25} {:<15}"
        separator_format = "{:<" + str(max_job_id_length) + "} {:<25} {:<15}"
        row_format = "{:<" + str(max_job_id_length) + "} {:<25} {:<15}"

        print(f"{message}:\n")
        print(header_format.format(*headers))

        separators = ["-" * len(header) for header in headers]
        print(separator_format.format(*separators))

        for job_id, timestamp, status in data:
            print(row_format.format(job_id, timestamp, status))


def _job_table_jupyter(data: list[str], message: Optional[str] = None, align: str = "right"):
    clear_output(wait=True)

    html = """<h3>Quantum Jobs</h3><table><tr>
    <th style='text-align:left'>qBraid ID</th>
    <th style='text-align:left'>Submitted</th>
    <th style='text-align:left'>Status</th></tr>
    """

    for job_id, timestamp, status_str in data:
        if status_str == "COMPLETED":
            color = "green"
        elif status_str == "FAILED":
            color = "red"
        elif status_str in [
            "INITIALIZING",
            "INITIALIZED",
            "CREATED",
            "QUEUED",
            "VALIDATING",
            "RUNNING",
        ]:
            color = "blue"
        else:
            color = "grey"

        status = f"<span style='color:{color}'>{status_str}</span>"

        html += f"""<tr>
        <td style='text-align:left'>{job_id}</td>
        <td style='text-align:left'>{timestamp}</td>
        <td style='text-align:left'>{status}</td></tr>
        """

    html += f"<tr><td colspan='4'; style='text-align:{align}'>{message}</td></tr>"

    html += "</table>"

    return display(HTML(html))


def display_jobs_from_data(job_data: list[list[str]], message: Optional[str] = None):
    """Displays a list of quantum jobs submitted by user, tabulated by job ID,
    the date/time it was submitted, and status."""
    align = "center" if len(job_data) == 0 else "right"

    if not message:
        num_jobs = len(job_data)
        if num_jobs == 0:
            message = "No jobs found matching criteria."
        else:
            message = f"Displaying {num_jobs} job{'s' if num_jobs > 1 else ''}."

    if running_in_jupyter():
        return _job_table_jupyter(job_data, message, align=align)
    return _job_table_basic(job_data, message)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=unused-argument

"""
Module for Cirq custom gates to aid the transpiler and qasm parser

"""

import fractions

import numpy as np
from cirq import CircuitDiagramInfo, Gate, IdentityGate, TwoQubitDiagonalGate


class U2Gate(Gate):
    """A single qubit gate for rotations about the
    X+Z axis of the Bloch sphere.
    """

    def __init__(self, phi, lam):
        self._phi = float(phi)
        self._lam = float(lam)

        super()

    def _num_qubits_(self) -> int:
        return 1

    def _unitary_(self):
        isqrt2 = 1 / np.sqrt(2)
        phi = self._phi
        lam = self._lam

        return np.array(
            [
                [isqrt2, -np.exp(1j * lam) * isqrt2],
                [
                    np.exp(1j * phi) * isqrt2,
                    np.exp(1j * (phi + lam)) * isqrt2,
                ],
            ],
        )

    def _circuit_diagram_info_(self, args):
        cirq_phi = self._phi / np.pi
        cirq_lam = self._lam / np.pi
        return f"U2({cirq_phi}, {cirq_lam})"

    def __str__(self) -> str:
        cirq_phi = fractions.Fraction(self._phi / np.pi)
        cirq_lam = fractions.Fraction(self._lam / np.pi)
        gate_str = f"U2({cirq_phi},{cirq_lam})"
        frac_str = gate_str.replace("/", "π/")
        return frac_str.replace("1π", "π")


class U3Gate(Gate):
    """A single qubit gate for generic rotations
    given 3 Euler angles.
    """

    def __init__(self, theta, phi, lam):
        self._theta = float(theta)
        self._phi = float(phi)
        self._lam = float(lam)

        super()

    def _num_qubits_(self) -> int:
        return 1

    def _unitary_(self):
        cos = np.cos(self._theta / 2)
        sin = np.sin(self._theta / 2)
        phi = self._phi
        lam = self._lam

        return np.array(
            [
                [cos, -np.exp(complex(0, lam)) * sin],
                [
                    np.exp(complex(0, phi)) * sin,
                    np.exp(complex(0, phi + lam)) * cos,
                ],
            ]
        )

    def _circuit_diagram_info_(self, args):
        cirq_theta = self._theta / np.pi
        cirq_phi = self._phi / np.pi
        cirq_lam = self._lam / np.pi
        return f"U3({cirq_theta}, {cirq_phi}, {cirq_lam})"

    def __str__(self) -> str:
        cirq_theta = fractions.Fraction(self._theta / np.pi)
        cirq_phi = fractions.Fraction(self._phi / np.pi)
        cirq_lam = fractions.Fraction(self._lam / np.pi)
        gate_str = f"U3({cirq_theta},{cirq_phi},{cirq_lam})"
        frac_str = gate_str.replace("/", "π/")
        return frac_str.replace("1π", "π")


class RZZGate(Gate):
    """A two qubit gate for rotations about ZZ."""

    def __init__(self, rads):
        self._rads = rads

        super()

    def _num_qubits_(self) -> int:
        return 2

    def _unitary_(self):
        itheta2 = 1j * self._rads * np.pi / 2
        matrix = np.array(
            [
                [np.exp(-itheta2), 0, 0, 0],
                [0, np.exp(itheta2), 0, 0],
                [0, 0, np.exp(itheta2), 0],
                [0, 0, 0, np.exp(-itheta2)],
            ],
        )
        return np.round(matrix, decimals=15)

    def _circuit_diagram_info_(self, args):
        rounded_theta = np.array(self._rads)
        if args.precision is not None:
            rounded_theta = rounded_theta.round(args.precision)
        gate_str = f"RZZ({rounded_theta})"
        return CircuitDiagramInfo((gate_str, gate_str))


def rzz(rads: float | int) -> Gate:
    """Returns custom cirq RZZ gate given rotation angle in radians."""
    if rads == 0:
        return IdentityGate(2)
    if rads == 2:
        return TwoQubitDiagonalGate([np.pi] * 4)

    # gate with the matrix exp(-i Z⊗Z rads)
    # return ZZPowGate(exponent=2 * rads / np.pi, global_shift=-0.5)
    return RZZGate(rads)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating dictionary of all braket gates for testing/benchmarking

"""

import string
from typing import Optional

import numpy as np
import scipy
from braket.circuits import Gate
from braket.circuits.gates import (
    CY,
    CZ,
    XX,
    XY,
    YY,
    ZZ,
    CCNot,
    CNot,
    CPhaseShift,
    H,
    I,
    ISwap,
    PhaseShift,
    PSwap,
    Rx,
    Ry,
    Rz,
    S,
    Si,
    Swap,
    T,
    Ti,
    V,
    Vi,
    X,
    Y,
    Z,
)

braket_gates = {
    # one-qubit, zero parameter
    "H": H,
    "X": X,
    "Y": Y,
    "Z": Z,
    "S": S,
    "Sdg": Si,
    "T": T,
    "Tdg": Ti,
    "I": I,
    "SX": V,
    "SXdg": Vi,
    # one-qubit, one parameter
    "Phase": PhaseShift,
    "RX": Rx,
    "RY": Ry,
    "RZ": Rz,
    "U1": PhaseShift,
    # two-qubit, zero parameter
    # 'CH':BraketGate.,
    "CX": CNot,
    "Swap": Swap,
    "iSwap": ISwap,
    # 'CSX':BraketGate.,
    # 'DCX': BraketGate.,
    "CY": CY,
    "CZ": CZ,
    # two-qubit, one parameter
    "RXX": XX,
    "RXY": XY,
    "RYY": YY,
    "RZZ": ZZ,
    "pSwap": PSwap,
    "CPhase": CPhaseShift,
    # multi-qubit
    "CCX": CCNot,
    # unitary
    # "Unitary": Unitary,
}


def generate_params(varnames: list[str], seed: Optional[int] = None):
    """Returns a dictionary of random parameters for a given list of variable names"""
    if seed is not None:
        np.random.seed(seed)
    params = {}
    for v in varnames:
        if v.startswith("angle"):
            params[v] = np.random.rand() * 2 * np.pi
    return params


def get_braket_gates(seed: Optional[int] = None):
    """Returns a dictionary of all braket gates with random parameters"""
    braket_gates = {attr: None for attr in dir(Gate) if attr[0] in string.ascii_uppercase}
    for gate in ["C", "PulseGate", "GPhase"]:
        braket_gates.pop(gate, None)

    for gate in braket_gates:
        if gate == "Unitary":
            n = np.random.randint(1, 4)
            unitary = scipy.stats.unitary_group.rvs(2**n)
            braket_gates[gate] = getattr(Gate, gate)(matrix=unitary)
        else:
            varnames = [v for v in getattr(Gate, gate).__init__.__code__.co_varnames if v != "self"]
            params = generate_params(varnames, seed=seed)
            braket_gates[gate] = getattr(Gate, gate)(**params)

    return {k: v for k, v in braket_gates.items() if v is not None}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating random IonQ programs

"""
from typing import Optional

import numpy as np

from qbraid.programs.gate_model.ionq import IonQProgram
from qbraid.programs.typer import IonQDict
from qbraid.transpiler import transpile

from .qasm3_random import qasm3_random_from_gates


def create_gateset_ionq(max_operands: int) -> np.ndarray:
    """Gets QASM for IonQ gateset with max_operands."""
    q1_gates: list[tuple[str, int, int]] = [
        ("x", 1, 0),
        ("y", 1, 0),
        ("z", 1, 0),
        ("h", 1, 0),
        ("s", 1, 0),
        ("t", 1, 0),
        ("v", 1, 0),
        ("si", 1, 0),
        ("ti", 1, 0),
        ("vi", 1, 0),
        ("rx", 1, 1),
        ("ry", 1, 1),
        ("rz", 1, 1),
    ]

    q2_gates: list[tuple[str, int, int]] = [
        ("cx", 2, 0),
        ("cy", 2, 0),
        ("cz", 2, 0),
        ("ch", 2, 0),
        ("crx", 2, 1),
        ("cry", 2, 1),
        ("crz", 2, 1),
        ("swap", 2, 0),
    ]

    q3_gates: list[tuple[str, int, int]] = [("ccnot", 3, 0)]

    gates = q1_gates.copy()

    if max_operands >= 2:
        gates.extend(q2_gates)
    if max_operands >= 3:
        gates.extend(q3_gates)

    gates_array = np.array(
        gates, dtype=[("gate", object), ("num_qubits", np.int64), ("num_params", np.int64)]
    )
    return gates_array


def ionq_random(
    num_qubits: Optional[int] = None,
    depth: Optional[int] = None,
    max_operands: Optional[int] = None,
    seed: Optional[int] = None,
) -> IonQDict:
    """Generate random OpenQASM 3 program.

    Args:
        num_qubits (int): Number of quantum wires.
        depth (int): Layers of operations (i.e., critical path length).
        max_operands (int): Maximum size of gate for each operation.
        seed (int): Seed for random number generator.
        measure (bool): Whether to include measurement gates.

    Raises:
        ValueError: When invalid random circuit options are given,
            or fails to generate valid IonQDict
        QbraidError: When failed to create random OpenQASM 3 program.

    Returns:
        IonQDict: Random IonQ program
    """
    qasm3_program = qasm3_random_from_gates(
        create_gateset_ionq, num_qubits, depth, max_operands, seed, False
    )

    ionq_dict = transpile(qasm3_program, "ionq")

    ionq_program = IonQProgram(ionq_dict)

    ionq_program.validate_for_gateset()

    return ionq_dict



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
CUDA-Q conversions

.. currentmodule:: qbraid.transpiler.conversions.cudaq

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   cudaq_to_qasm2
   cudaq_to_pyqir

"""
from .cudaq_extras import cudaq_to_pyqir
from .cudaq_to_qasm2 import cudaq_to_qasm2

__all__ = ["cudaq_to_qasm2", "cudaq_to_pyqir"]



# Copyright (C) 2024 qBraid
# Copyright (C) 2024 QuEra Computing Inc.
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for animating QuEra simulation results using Flair Visual.

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from .exceptions import VisualizationError

if TYPE_CHECKING:
    import flair_visual.animation.runtime.qpustate
    import matplotlib.animation


def animate_qpu_state(  # pylint: disable=too-many-arguments
    state: Union[flair_visual.animation.runtime.qpustate.AnimateQPUState, dict],
    dilation_rate: float = 0.05,
    fps: int = 30,
    gate_display_dilation: float = 1.0,
    save_mpeg: bool = False,
    filename: str = "vqpu_animation",
    start_block: int = 0,
    n_blocks: Optional[int] = None,
    **kwargs,
) -> matplotlib.animation.FuncAnimation:
    """Animates the QPU state.

    Args:
        state (Union[AnimateQPUState, dict]): The state of the QPU atoms to animate.
        dilation_rate (float): Conversion factor from QPU time to animation time units.
            When `dilation_rate=1.0`, 1 μs of QPU execution time corresponds to 1 second
            of animation time.
        fps (int): Frames per second. Defaults to 30.
        gate_display_dilation (float, optional): Relative dilation rate for gate events.
            Defaults to 1. Higher values will display gate events for a longer period.
        save_mpeg (bool): Whether to save the animation as an MPEG file. Defaults to False.
        filename (str): The name of the '.mpeg' file to save. Defaults to "vqpu_animation".
            Ignored if `save_mpeg` is False.
        start_block (int): The starting block for the animation. Defaults to 0.
        n_blocks (int, optional): The number of blocks to animate. Defaults to None. If None,
            all blocks after `start_block` will be animated.

    Returns:
        matplotlib.animation.FuncAnimation: The generated animation object for the QPU state.

    Raises:
        VisualizationError: If an error occurs during the animation process.
    """
    # pylint: disable=import-outside-toplevel
    from flair_visual.animation.animate import animate_qpu_state as flair_animate_qpu_state
    from flair_visual.animation.runtime.qpustate import AnimateQPUState

    # pylint: enable=import-outside-toplevel

    if isinstance(state, dict):
        state = AnimateQPUState.from_json(state)

    try:
        animation = flair_animate_qpu_state(
            state=state,
            dilation_rate=dilation_rate,
            fps=fps,
            gate_display_dilation=gate_display_dilation,
            save_mpeg=save_mpeg,
            filename=filename,
            start_block=start_block,
            n_blocks=n_blocks,
            **kwargs,
        )
    except Exception as err:  # pylint: disable=broad-exception-caught
        raise VisualizationError from err

    return animation



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for managing conversion configurations for quantum runtime.

"""
from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, Optional, Union

from qbraid.programs.spec import ProgramSpec

from .graph import ConversionGraph, parse_conversion_path

if TYPE_CHECKING:
    import rustworkx as rx


@dataclass
class ConversionScheme:
    """
    A data class for managing conversion configurations for quantum device operations.

    Attributes:
        conversion_graph (Optional[ConversionGraph]): Graph coordinating conversions between
            different quantum software program types. If None, the default qBraid graph is used.
        max_path_attempts (int): The maximum number of conversion paths to attempt before
            raising an exception. Defaults to 3.
        max_path_depth (Optional[int]): The maximum depth of conversions within a given path to
            allow. A depth of 2 would allow a conversion path like ['cirq' -> 'qasm2' -> 'qiskit'].
            Defaults to None, meaning no limit.
        extra_kwargs (dict[str, Any]): A dictionary to hold any additional keyword arguments that
            users want to pass to the transpile function at runtime.

    Methods:
        to_dict: Converts the conversion scheme to a flat dictionary suitable for passing as kwargs.
        update_values: Dynamically updates the values of the instance's attributes.
    """

    conversion_graph: Optional[ConversionGraph] = None
    max_path_attempts: int = 3
    max_path_depth: Optional[int] = None
    extra_kwargs: dict[str, Any] = field(default_factory=dict)

    def __str__(self):
        kwargs_str = ", ".join(f"{key}={value}" for key, value in self.extra_kwargs.items())
        return (
            f"ConversionScheme(conversion_graph={self.conversion_graph}, "
            f"max_path_attempts={self.max_path_attempts}, "
            f"max_path_depth={self.max_path_depth}, "
            f"{kwargs_str})"
        )

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the ConversionScheme fields to a flat dictionary suitable for passing as kwargs.

        Returns:
            A dictionary with all fields ready to be passed as keyword arguments,
            including nested extra_kwargs.
        """
        scheme = asdict(self)
        scheme.update(scheme.pop("extra_kwargs", {}))
        scheme.update({"conversion_graph": self.conversion_graph})
        return scheme

    def update_values(self, **kwargs) -> None:
        """
        Updates the attributes of the conversion scheme with new values provided
        as keyword arguments.

        Args:
            **kwargs: Arbitrary keyword arguments containing attribute names and their new values.

        Raises:
            AttributeError: If a provided attribute name does not exist.
        """
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError(f"{key} is not a valid attribute of ConversionScheme")

    @staticmethod
    def find_nodes_reachable_within_max_edges(
        graph: rx.PyDiGraph,
        target_nodes: Union[list[str], set[str]],
        max_edges: Optional[int] = None,
    ) -> set[str]:
        """Find all nodes reachable from a target node within a specified number of edges.

        Args:
            graph (rx.PyDiGraph): The graph to search.
            target_nodes (list[str]): The target nodes from which to search.
            max_edges (int, optional): The maximum number of edges to traverse.

        Returns:
            set[str]: The set of nodes reachable from the target nodes within the specified
                number of edges.

        Raises:
            ValueError: If the target node is not found in the graph,
                or if the maximum number of edges is negative.
        """
        if max_edges is None:
            max_edges = graph.num_edges()
        elif max_edges < 0:
            raise ValueError("The maximum number of edges must be a non-negative integer.")

        graph_nodes = graph.nodes()
        node_to_index = {node: i for i, node in enumerate(graph_nodes)}
        target_indices = set()

        for target_node in set(target_nodes):
            if target_node not in node_to_index:
                raise ValueError(f"Target node '{target_node}' not found in the graph.")
            target_indices.add(node_to_index[target_node])

        reachable_nodes = set(target_indices)

        for _ in range(max_edges):
            new_nodes = set()
            for node in reachable_nodes:
                preds = graph.predecessors(node)
                preds_indices = [node_to_index[pred] for pred in preds]
                new_nodes.update(preds_indices)
            if not new_nodes.difference(reachable_nodes):
                break
            reachable_nodes.update(new_nodes)

        return {graph_nodes[i] for i in reachable_nodes}

    @staticmethod
    def prune_graph_to_target_paths(
        graph: ConversionGraph, target_nodes: list[str], n_steps: Optional[int]
    ) -> ConversionGraph:
        """
        Prune edges that do not contribute to paths within n steps of any of the target nodes.

        Args:
            graph (ConversionGraph): The graph to prune
            target_nodes (List[str]): The list of node indices to center the pruning around
            n_steps (int): Number of steps to consider from each target node

        Returns:
            ConversionGraph: The pruned graph
        """
        graph = graph.copy()

        all_nodes: set[str] = set(graph.nodes())
        target_set: set[str] = set(target_nodes)
        sources: set[str] = all_nodes - target_set
        all_used_paths: set[tuple[str, str]] = set()

        # Collect all paths within n_steps for each target node
        for target_node in target_nodes:
            for source in sources:
                if graph.has_path(source, target_node):
                    for path_str in graph.all_paths(source, target_node):
                        path_tuple_lst = parse_conversion_path(path_str)
                        if n_steps is None or len(path_tuple_lst) <= n_steps:
                            all_used_paths.update(path_tuple_lst)

        # Create a mapping from node IDs to aliases
        node_id_to_alias = {value: key for key, value in graph._node_alias_id_map.items()}

        # Prune edges not in used paths
        for edge in graph.edge_list():
            src_node_id, target_node_id = edge
            src_node_alias = node_id_to_alias[src_node_id]
            target_node_alias = node_id_to_alias[target_node_id]

            if (src_node_alias, target_node_alias) not in all_used_paths:
                graph.remove_conversion(src_node_alias, target_node_alias)

        return graph

    def reset_graph(self, include_isolated: bool = True) -> None:
        """Reset the conversion graph to the default qBraid graph."""
        self.update_values(conversion_graph=ConversionGraph(include_isolated=include_isolated))

    def update_graph_for_target(self, target_spec: Union[ProgramSpec, list[ProgramSpec]]) -> None:
        """Update the conversion graph to include only nodes with paths to the target node(s), and
        remove all conversions that do not end in the target node(s)."""
        graph = (
            self.conversion_graph.copy()
            if self.conversion_graph
            else ConversionGraph(include_isolated=True)
        )

        target_nodes = {
            spec.alias for spec in (target_spec if isinstance(target_spec, list) else [target_spec])
        }

        nodes = self.find_nodes_reachable_within_max_edges(graph, target_nodes, self.max_path_depth)

        conversions = [conv for conv in graph.conversions() if conv.source not in target_nodes]

        updated_graph = ConversionGraph(
            conversions=conversions,
            require_native=graph.require_native,
            include_isolated=graph._include_isolated,
            edge_bias=graph.edge_bias,
            nodes=nodes,
        )

        pruned_graph = self.prune_graph_to_target_paths(
            updated_graph, target_nodes, self.max_path_depth
        )

        self.update_values(conversion_graph=pruned_graph)



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for Equal1 simulator runtime through native provider.

"""
import pytest

from qbraid.runtime.native.result import Equal1SimulatorResultData
from qbraid.runtime.schemas.experiment import Equal1SimulationMetadata


def test_equal1_simulator_result_data():
    """Test the Equal1SimulatorResultData class."""
    compiled_output = (
        "T1BFTlFBU00gMi4wOwppbmNsdWRlICJxZWxpYjEuaW5jIjsKcXJlZyBxWzZdOwpjcmVnIG1lYXNbMl07"
        "CnJ6KHBpLzIpIHFbMl07CnJ4KHBpLzIpIHFbMl07CnJ6KHBpLzIpIHFbMl07CnJ6KHBpLzIpIHFbM107"
        "CnJ4KHBpLzIpIHFbM107CnJ6KHBpKSBxWzNdOwpjeiBxWzJdLHFbM107CnJ4KHBpLzIpIHFbM107"
        "CnJ6KHBpLzIpIHFbM107CmJhcnJpZXIgcVsyXSxxWzNdOwptZWFzdXJlIHFbMl0gLT4gbWVhc1swXTsK"
        "bWVhc3VyZSBxWzNdIC0+IG1lYXNbMV07"
    )
    qasm_expected = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[6];
creg meas[2];
rz(pi/2) q[2];
rx(pi/2) q[2];
rz(pi/2) q[2];
rz(pi/2) q[3];
rx(pi/2) q[3];
rz(pi) q[3];
cz q[2],q[3];
rx(pi/2) q[3];
rz(pi/2) q[3];
barrier q[2],q[3];
measure q[2] -> meas[0];
measure q[3] -> meas[1];"""
    result_data = Equal1SimulatorResultData(compiled_output)
    assert result_data.compiled_output == qasm_expected


@pytest.fixture
def equal1_partial_data():
    """Fixture providing partial Equal1 simulation metadata fields."""
    return {
        "compiledOutput": (
            "T1BFTlFBU00gMi4wOwppbmNsdWRlICJxZWxpYjEuaW5jIjsKcXJlZyBxWzZdOwpjcmVnIG1lYXNbMl07"
        ),
        "executionMode": "simulation",
        "deviceName": "bell1-6-lin",
    }


@pytest.fixture
def equal1_full_data(equal1_partial_data):
    """Fixture providing complete Equal1 simulation metadata example data."""
    return {
        **equal1_partial_data,
        "simulationPlatform": "CPU",
        "simulationMode": None,
        "executionOptions": {"force_bypass_transpilation": 0, "optimization_level": 3},
    }


@pytest.fixture
def equal1_data_with_qasm(equal1_partial_data):
    """Fixture providing Equal1 data with QASM and measurement counts."""
    qasm = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
creg c[2];
h q[0];
cx q[0],q[1];
measure q[0] -> c[0];
measure q[1] -> c[1];
"""
    return {
        **equal1_partial_data,
        "openQasm": qasm,
        "measurementCounts": {"00": 100, "11": 100},
    }


def test_equal1_simulation_metadata(equal1_full_data):
    """Test the Equal1SimulationMetadata class with complete example data."""
    # Create metadata instance
    metadata = Equal1SimulationMetadata(**equal1_full_data)

    # Test that all fields are correctly set
    assert metadata.compiled_output == equal1_full_data["compiledOutput"]
    assert metadata.execution_mode == equal1_full_data["executionMode"]
    assert metadata.device_name == equal1_full_data["deviceName"]
    assert metadata.simulation_platform == equal1_full_data["simulationPlatform"]
    assert metadata.simulation_mode == equal1_full_data["simulationMode"]
    assert metadata.execution_options == equal1_full_data["executionOptions"]

    # Test that the base class fields are accessible (inherited from GateModelExperimentMetadata)
    assert metadata.measurement_counts is None
    assert metadata.measurements is None
    assert metadata.qasm is None
    assert metadata.num_qubits is None
    assert metadata.gate_depth is None


def test_equal1_simulation_metadata_partial(equal1_partial_data):
    """Test Equal1SimulationMetadata with partial data."""
    metadata = Equal1SimulationMetadata(**equal1_partial_data)

    # Test that provided fields are set
    assert metadata.compiled_output == equal1_partial_data["compiledOutput"]
    assert metadata.execution_mode == equal1_partial_data["executionMode"]
    assert metadata.device_name == equal1_partial_data["deviceName"]

    # Test that optional fields are None
    assert metadata.simulation_platform is None
    assert metadata.simulation_mode is None
    assert metadata.execution_options is None


def test_equal1_simulation_metadata_empty():
    """Test Equal1SimulationMetadata with empty data."""
    metadata = Equal1SimulationMetadata()

    # Test that all fields are None when no data is provided
    assert metadata.compiled_output is None
    assert metadata.execution_mode is None
    assert metadata.device_name is None
    assert metadata.simulation_platform is None
    assert metadata.simulation_mode is None
    assert metadata.execution_options is None


def test_equal1_simulation_metadata_with_qasm(equal1_data_with_qasm):
    """Test Equal1SimulationMetadata with QASM data to test inheritance."""
    metadata = Equal1SimulationMetadata(**equal1_data_with_qasm)

    # Test Equal1 specific fields
    assert metadata.compiled_output == equal1_data_with_qasm["compiledOutput"]
    assert metadata.execution_mode == equal1_data_with_qasm["executionMode"]
    assert metadata.device_name == equal1_data_with_qasm["deviceName"]

    # Test inherited fields
    assert metadata.qasm == equal1_data_with_qasm["openQasm"]
    assert metadata.measurement_counts is not None
    assert metadata.measurement_counts["00"] == 100
    assert metadata.measurement_counts["11"] == 100



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining qBraid runtime job schemas.

"""
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, field_serializer

from .base import USD, Credits


class DevicePricing(BaseModel):
    """Represents pricing information for a quantum device.

    Attributes:
        perTask (USD): The cost per task in USD.
        perShot (USD): The cost per shot in USD.
        perMinute (USD): The cost per minute in USD.
    """

    model_config = ConfigDict(frozen=True)

    perTask: USD
    perShot: USD
    perMinute: USD

    @field_serializer("perTask", "perShot", "perMinute")
    def serialize_credits(self, value: USD) -> Credits:
        """Serialize USD fields into Credits objects."""
        return Credits(value.to_credits())


class DeviceData(BaseModel):
    """Represents metadata and capabilities of a quantum device.

    Attributes:
        provider (str): The entity that manufactures the quantum hardware or maintains the
            simulator software.
        vendor (str): The entity that hosts or provides access to the quantum device for end users.
        name (str): The name of the quantum device.
        paradigm (str): The quantum computing paradigm (e.g., gate-model, AHS).
        status (str): The current status of the device (e.g., ONLINE, OFFLINE).
        is_available (bool): Indicates whether the device is available for jobs.
        queue_depth (int, optional): The depth of the job queue, or None if not applicable.
        device_type (str): The type of device (e.g., Simulator, QPU).
        num_qubits (int): The number of qubits supported by the device.
        run_input_types (list[str]): The software packages / program type alises that can be
            used to specify quantum programs in jobs submitted to the device (e.g. ['qasm2']).
        device_id (str): The qBraid-specific device identifier.
        noise_models (list[str], optional): A list of supported noise models. Defaults to None.
        pricing (DevicePricing): The pricing structure for using the device, in qBraid credits.
    """

    model_config = ConfigDict(frozen=True)

    provider: str
    vendor: str
    name: str
    paradigm: str
    status: str

    is_available: bool = Field(alias="isAvailable")
    queue_depth: Optional[int] = Field(None, alias="queueDepth")

    device_type: str = Field(alias="type")
    num_qubits: Optional[int] = Field(alias="numberQubits")
    run_input_types: list[str] = Field(alias="runInputTypes")
    device_id: str = Field(alias="qbraid_id")

    noise_models: Optional[list[str]] = Field(None, alias="noiseModels")
    pricing: Optional[DevicePricing] = None



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Test circuit equality helper functions

"""
import numpy as np
import pytest

from qbraid.interface.circuit_equality import assert_allclose_up_to_global_phase, match_global_phase


def test_match_global_phase_basic1():
    """Test matching global phase of two arrays"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([1 - 1j, 2 - 2j])
    a_prime, b_prime = match_global_phase(a, b)
    np.testing.assert_allclose(a_prime, b_prime, atol=1e-10)


def test_match_global_phase_basic2():
    """Test matching global phase of two arrays"""
    a = np.array([[5, 4], [3, -2]])
    b = np.array([[0.000001, 0], [0, 1j]])
    c, d = match_global_phase(a, b)
    np.testing.assert_allclose(c, -a, atol=1e-10)
    np.testing.assert_allclose(d, b * -1j, atol=1e-10)


def test_match_global_phase_shape_mismatch1():
    """Test matching global phase of two arrays with shape mismatch"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([[1 - 1j, 2 - 2j]])
    a_prime, b_prime = match_global_phase(a, b)
    np.testing.assert_allclose(a, a_prime, atol=1e-10)
    np.testing.assert_allclose(b, b_prime, atol=1e-10)


def test_match_global_phase_shape_mismatch2():
    """Test matching global phase of two arrays with shape mismatch"""
    a = np.array([1])
    b = np.array([1, 2])
    c, d = match_global_phase(a, b)
    assert c.shape == a.shape
    assert d.shape == b.shape
    assert c is not a
    assert d is not b
    assert np.allclose(c, a)
    assert np.allclose(d, b)

    a = np.array([])
    b = np.array([])
    c, d = match_global_phase(a, b)
    assert c.shape == a.shape
    assert d.shape == b.shape
    assert c is not a
    assert d is not b
    assert np.allclose(c, a)
    assert np.allclose(d, b)


def test_match_global_phase_zeros():
    """Test matching global phase of two arrays where one is all zeros"""
    z = np.array([[0, 0], [0, 0]])
    b = np.array([[1, 1], [1, 1]])

    z1, b1 = match_global_phase(z, b)
    np.testing.assert_allclose(z, z1, atol=1e-10)
    np.testing.assert_allclose(b, b1, atol=1e-10)

    z2, b2 = match_global_phase(z, b)
    np.testing.assert_allclose(z, z2, atol=1e-10)
    np.testing.assert_allclose(b, b2, atol=1e-10)

    z3, z4 = match_global_phase(z, z)
    np.testing.assert_allclose(z, z3, atol=1e-10)
    np.testing.assert_allclose(z, z4, atol=1e-10)


def test_match_global_phase_invariance_to_phase_and_sign():
    """Test match global phase function remains invariant (or consistent) when the
    input arrays undergo transformations related to their phase or sign changes."""
    a = np.array([[1, 1.1], [-1.3, np.pi]])
    a2, _ = match_global_phase(a, a)
    a3, _ = match_global_phase(a * 1j, a * 1j)
    a4, _ = match_global_phase(-a, -a)
    a5, _ = match_global_phase(a * -1j, a * -1j)

    assert np.all(a2 == a)
    assert np.all(a3 == a)
    assert np.all(a4 == a)
    assert np.all(a5 == a)


def test_assert_allclose_up_to_global_phase_basic():
    """Test assert_allclose_up_to_global_phase with two arrays"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([1 - 1j, 2 - 2j])
    assert_allclose_up_to_global_phase(a, b, atol=1e-10)


def test_assert_allclose_up_to_global_phase_fail():
    """Test assert_allclose_up_to_global_phase with two arrays that are not close"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([10 + 10j, 20 + 20j])
    with pytest.raises(AssertionError):
        assert_allclose_up_to_global_phase(a, b, atol=1e-10)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating random Cirq circuits

"""
import numpy as np
from cirq import Circuit
from cirq.testing import random_circuit

from qbraid.exceptions import QbraidError


def cirq_random(num_qubits: int, depth: int, **kwargs) -> Circuit:
    """Generate random circuit of arbitrary size and form.

    Args:
        num_qubits (int): number of quantum wires
        depth (int): layers of operations (i.e. critical path length)

    Raises:
        QbraidError: When invalid cirq random circuit options given

    Returns:
        Cirq random circuit

    """
    if "random_state" not in kwargs:
        kwargs["random_state"] = np.random.randint(1, 11)

    if "op_density" not in kwargs:
        kwargs["op_density"] = 1

    try:
        return random_circuit(num_qubits, n_moments=depth, **kwargs)
    except (ValueError, TypeError) as err:
        raise QbraidError("Failed to create Cirq random circuit") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for matplotlib animations using Flair Visual

"""
import logging
from unittest.mock import MagicMock, patch

import pytest

from qbraid.visualization.exceptions import VisualizationError
from qbraid.visualization.flair_animations import animate_qpu_state

try:
    from flair_visual.animation.runtime.qpustate import AnimateQPUState

    flair_visual_installed = True
except ImportError as err:
    flair_visual_installed = False

    logging.warning("flair_animations tests will be skipped: %s", err)

pytestmark = pytest.mark.skipif(
    flair_visual_installed is False, reason="flair-visual not installed"
)


@pytest.fixture
def mock_state():
    """Fixture for an empty mock AnimateQPUState object."""
    return AnimateQPUState()


@pytest.fixture
def mock_state_from_json():
    """Fixture for a mock AnimateQPUState created from JSON."""
    state_data = {
        "block_durations": [],
        "gate_events": [],
        "qpu_fov": {"xmin": None, "xmax": None, "ymin": None, "ymax": None},
        "atoms": [],
        "slm_zone": [],
        "aod_moves": [],
    }
    return AnimateQPUState.from_json(state_data)


@patch("flair_visual.animation.animate.animate_qpu_state")
def test_animate_qpu_state_with_animate_qpustate(mock_animate_qpu_state, mock_state):
    """Test the animate_qpu_state function with an AnimateQPUState object."""
    mock_animation = MagicMock()
    mock_animate_qpu_state.return_value = mock_animation

    result = animate_qpu_state(state=mock_state, dilation_rate=0.1, fps=24, save_mpeg=False)

    mock_animate_qpu_state.assert_called_once_with(
        state=mock_state,
        dilation_rate=0.1,
        fps=24,
        gate_display_dilation=1.0,
        save_mpeg=False,
        filename="vqpu_animation",
        start_block=0,
        n_blocks=None,
    )

    assert result == mock_animation


@patch("flair_visual.animation.animate.animate_qpu_state")
def test_animate_qpu_state_with_json(mock_animate_qpu_state, mock_state_from_json):
    """Test the animate_qpu_state function with a JSON dict input."""
    mock_animation = MagicMock()
    mock_animate_qpu_state.return_value = mock_animation

    state_data = {
        "block_durations": [],
        "gate_events": [],
        "qpu_fov": {"xmin": None, "xmax": None, "ymin": None, "ymax": None},
        "atoms": [],
        "slm_zone": [],
        "aod_moves": [],
    }
    result = animate_qpu_state(
        state=state_data, dilation_rate=0.1, fps=24, save_mpeg=True, filename="test_animation"
    )

    mock_animate_qpu_state.assert_called_once_with(
        state=mock_state_from_json,
        dilation_rate=0.1,
        fps=24,
        gate_display_dilation=1.0,
        save_mpeg=True,
        filename="test_animation",
        start_block=0,
        n_blocks=None,
    )
    assert result == mock_animation


@patch("flair_visual.animation.animate.animate_qpu_state")
def test_animate_qpu_state_raises_visualization_error(mock_animate_qpu_state, mock_state):
    """Test that a VisualizationError is raised when an exception occurs."""
    mock_animate_qpu_state.side_effect = ValueError("Test error")

    with pytest.raises(VisualizationError):
        animate_qpu_state(state=mock_state)


@patch("flair_visual.animation.animate.animate_qpu_state")
def test_animate_qpu_state_with_extra_kwargs(mock_animate_qpu_state, mock_state):
    """Test the animate_qpu_state function with additional kwargs."""
    mock_animation = MagicMock()
    mock_animate_qpu_state.return_value = mock_animation

    result = animate_qpu_state(state=mock_state, dilation_rate=0.2, fps=60, extra_arg=True)

    mock_animate_qpu_state.assert_called_once_with(
        state=mock_state,
        dilation_rate=0.2,
        fps=60,
        gate_display_dilation=1.0,
        save_mpeg=False,
        filename="vqpu_animation",
        start_block=0,
        n_blocks=None,
        extra_arg=True,
    )
    assert result == mock_animation



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=arguments-differ

"""
Module defining QbraidDevice class

"""
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Optional, Union

from qbraid_core.decimal import Credits
from qbraid_core.services.quantum import QuantumClient, QuantumServiceRequestError

from qbraid._entrypoints import get_entrypoints
from qbraid._logging import logger
from qbraid.programs import ExperimentType, ProgramSpec, get_program_type_alias, load_program
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.runtime.exceptions import QbraidRuntimeError
from qbraid.runtime.noise import NoiseModel
from qbraid.runtime.schemas.experiment import QuboSolveParams
from qbraid.runtime.schemas.job import RuntimeJobModel
from qbraid.transpiler import ConversionGraph, transpile

from .job import QbraidJob

if TYPE_CHECKING:
    import qbraid_core.services.quantum

    import qbraid.programs
    import qbraid.runtime
    import qbraid.transpiler


class QbraidDevice(QuantumDevice):
    """Class to represent a qBraid device."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        client: Optional[qbraid_core.services.quantum.QuantumClient] = None,
        **kwargs,
    ):
        """Create a new QbraidDevice object."""
        super().__init__(profile=profile, **kwargs)
        self._client = client or QuantumClient()

    @property
    def client(self) -> QuantumClient:
        """Return the QuantumClient object."""
        return self._client

    def __str__(self):
        """String representation of the QbraidDevice object."""
        return f"{self.__class__.__name__}('{self.id}')"

    def status(self) -> qbraid.runtime.DeviceStatus:
        """Return device status."""
        device_data = self.client.get_device(self.id)
        status: Optional[str] = device_data.get("status")
        if not status:
            raise QbraidRuntimeError("Failed to retrieve device status")
        return DeviceStatus(status.lower())

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the backend"""
        device_data = self.client.get_device(self.id)
        pending_jobs = device_data.get("pendingJobs", 0)
        return int(pending_jobs)

    def submit(  # pylint: disable=too-many-arguments
        self,
        run_input: dict[str, Union[bytes, str]],
        memory: bool = False,
        preflight: bool = False,
        shots: Optional[int] = None,
        tags: Optional[dict[str, str]] = None,
        entrypoint: Optional[str] = None,
        noise_model: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        backend: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
        error_mitigation: Optional[dict[str, Any]] = None,
        runtime_options: Optional[dict[str, Any]] = None,
    ) -> qbraid.runtime.QbraidJob:
        """
        Creates a qBraid Quantum Job.

        Args:
            run_input (dict[str, Union[bytes, str]]): Dictionary containing
                QIR bytecode or OpenQASM string to be run on the device.
            memory (bool, optional): Whether to retain the individual shot results.
                Only applicable for certain devices. Defaults to False.
            preflight (bool, optional): Whether to run the job in preflight mode.
                Only applicable for certain devices. Defaults to False.
            shots (int, optional): The number of times to repeat the execution of the
                program. Default value varies by device.
            tags (dict, optional): A dictionary of tags to associate with the job.
            entrypoint (str, optional): Name of the entrypoint function to execute.
                Only applicable if run_input is a QIR module. Defaults to None.
            noise_model (str, optional): The noise model to apply to the job.
                Only applicable if device supports noisey simulation. Defaults to None.
            seed (int, optional): The seed to use for the random number generator.
                Only applicable for certain devices. Defaults to None.
            timeout (int, optional): The maximum time in seconds to wait for the job
                to complete after execution has started. Defaults to None.
            backend (str, optional): The backend to use for the job. Only applicable for
                simulator devices that support multiple backend executables. Defaults to None.
            params (dict, optional): Additional parameters to include in the job payload.
            error_mitigation (dict, optional): Dictionary containing error mitigation
                parameters. Only applicable for certain devices. Defaults to None.
            runtime_options (dict, optional): Additional runtime options to include in the
                job payload. Only applicable for certain devices. Defaults to None.

        Returns:
            QbraidJob: The job objects representing the submitted job.

        See Also: https://docs.qbraid.com/api-reference/api-reference/post-quantum-jobs

        """
        tags = tags or {}
        tags_json = json.dumps(tags)
        error_mitig_json = json.dumps(error_mitigation) if error_mitigation else None
        params_json = QuboSolveParams(**params).model_dump_json() if params else None
        runtime_options_json = json.dumps(runtime_options) if runtime_options else None

        payload = {
            "qbraidDeviceId": self.id,
            "tags": tags_json,
            "shots": shots,
            "seed": seed,
            "entrypoint": entrypoint,
            "timeout": timeout,
            "noiseModel": noise_model,
            "memory": memory,
            "preflight": preflight,
            "backend": backend,
            "params": params_json,
            "errorMitigation": error_mitig_json,
            **run_input,
        }

        # TODO: Move runtimeOptions to initial payload declaration once supported by qBraid API
        if runtime_options_json:
            payload["runtimeOptions"] = runtime_options_json  # pragma: no cover

        job_data = self.client.create_job(data=payload)

        payload.update(job_data)
        payload["tags"] = tags
        payload["params"] = params
        payload["errorMitigation"] = error_mitigation
        payload["runtimeOptions"] = runtime_options
        job_model = RuntimeJobModel.from_dict(payload)
        model_dump = job_model.model_dump(exclude={"metadata", "cost"})
        return QbraidJob(**model_dump, device=self, client=self.client)

    def try_extracting_info(self, func, error_message):
        """Try to extract information from a function/attribute,
        logging an error if it fails."""
        try:
            return func()
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.info("%s: %s. Field will be omitted in job metadata.", error_message, str(err))
            return None

    def _extract_qasm_rep(
        self, program: qbraid.programs.QPROGRAM, program_spec: ProgramSpec
    ) -> Optional[str]:
        """Populate the qasm info in the payload."""
        if program_spec.alias in ["qasm2", "qasm3"]:
            return program

        aux_graph = ConversionGraph()

        closest_qasm = aux_graph.closest_target(program_spec.alias, ["qasm2", "qasm3"])

        if closest_qasm is None:
            return None

        return transpile(program, closest_qasm, conversion_graph=aux_graph)

    def _construct_aux_payload(
        self, program: qbraid.programs.QPROGRAM, program_spec: Optional[ProgramSpec] = None
    ) -> dict[str, Union[int, str]]:
        """Construct auxiliary payload for the job submission."""
        aux_payload = {}

        if program_spec is None:
            program_alias = get_program_type_alias(program, safe=True)
            program_spec = ProgramSpec(type(program), alias=program_alias)

        if program_spec.native is False:
            return aux_payload

        qbraid_program = load_program(program)

        payload_key = {
            ExperimentType.GATE_MODEL: "circuitNumQubits",
            ExperimentType.ANNEALING: "numVariables",
            ExperimentType.AHS: "numAtoms",
        }

        num_required_qubits = self.try_extracting_info(
            lambda program=qbraid_program: program.num_qubits,
            "Error calculating circuit number of qubits.",
        )

        key = payload_key.get(program_spec.experiment_type)
        if num_required_qubits is not None and key:
            aux_payload[key] = num_required_qubits

        if program_spec.experiment_type != ExperimentType.GATE_MODEL:
            return aux_payload

        aux_payload["circuitDepth"] = self.try_extracting_info(
            lambda program=qbraid_program: program.depth, "Error calculating circuit depth."
        )
        aux_payload["openQasm"] = self.try_extracting_info(
            lambda program=program, program_spec=program_spec: self._extract_qasm_rep(
                program, program_spec
            ),
            "Error extracting OpenQASM string representation.",
        )

        return aux_payload

    def _resolve_noise_model(self, noise_model: Union[NoiseModel, str]) -> str:
        """Verify given noise model is supported by device and map to string representation."""
        if self.profile.noise_models is None:
            raise ValueError("Noise models are not supported by this device.")

        if isinstance(noise_model, NoiseModel):
            noise_model = noise_model.value
        elif not isinstance(noise_model, str):
            raise ValueError(
                f"Invalid type for noise model: {type(noise_model)}. Expected str or NoiseModel."
            )

        if noise_model not in self.profile.noise_models:
            raise ValueError(f"Noise model '{noise_model}' not supported by device.")

        return self.profile.noise_models.get(noise_model).name

    def _resolve_qubo_params(self, params: Union[QuboSolveParams, dict]) -> dict[str, Any]:
        """Resolve QUBO solve parameters to a dictionary."""
        if self.profile.experiment_type != ExperimentType.ANNEALING:
            raise ValueError("QUBO solve parameters are only supported for annealing devices.")

        if isinstance(params, QuboSolveParams):
            params = params.model_dump()
        elif not isinstance(params, dict):
            raise ValueError(
                f"Invalid type for QUBO solve parameters: {type(params)}. "
                "Expected dict or QuboSolveParams."
            )

        return params

    @staticmethod
    def _validate_run_input_payload(
        payload: Union[dict[str, Any], Any], target_spec: Optional[ProgramSpec]
    ) -> None:
        """Raises an exception if the transformed run input is not a dictionary."""
        if not isinstance(payload, dict):
            error_message = (
                "Run input transform failed{}. If the issue persists, "
                "please submit a bug report at https://github.com/qBraid/qBraid/issues."
            )

            if target_spec is None:
                error_message = error_message.format(
                    " due to missing target ProgramSpec. Ensure all required "
                    "dependency extras for this device are installed, and try again"
                )
            else:
                error_message = error_message.format(
                    ", likely due to corrupted target ProgramSpec. Use QbraidProvider.get_device() "
                    "to re-instantiate the device object, and try again"
                )

            raise QbraidRuntimeError(error_message)

    @staticmethod
    def _all_target_specs_native(program_spec: ProgramSpec | list[ProgramSpec] | None) -> bool:
        """Returns True if all of the given ProgramSpec's are natively supported by qBraid."""
        entrypoints = get_entrypoints("programs")
        specs = [program_spec] if isinstance(program_spec, ProgramSpec) else program_spec or []
        return all(spec.native and spec.alias in entrypoints for spec in specs)

    def run(
        self,
        run_input: Union[qbraid.programs.QPROGRAM, list[qbraid.programs.QPROGRAM]],
        shots: Optional[int] = None,
        tags: Optional[dict[str, str]] = None,
        **kwargs,
    ) -> Union[qbraid.runtime.QbraidJob, list[qbraid.runtime.QbraidJob]]:
        """
        Run a quantum job or a list of quantum jobs on this quantum device.

        Args:
            run_input (Union[QPROGRAM, list[QPROGRAM]]): A single quantum program
                or a list of quantum programs to run on the device.
            shots (optional, int): The number of times to repeat the execution of the
                program. Default value varies by device.
            tags (optional, dict): A dictionary of tags to associate with the job.
            **kwargs: Additional json data to include in the job submission payload.

        Returns:
            A QuantumJob object or a list of QuantumJob objects corresponding to the input.

        Raises:
            ValueError: If any protected dynamic parameters are specified in the kwargs.
        """
        dynamic_params = {
            "openQasm": None,
            "ionqCircuit": None,
            "bitcode": None,
            "problem": None,
            "numVariables": None,
            "numAtoms": None,
            "circuitNumQubits": None,
            "circuitDepth": None,
        }
        forbidden_keys = set(dynamic_params.keys())

        if any(key in kwargs for key in forbidden_keys):
            raise ValueError(
                f"You cannot specify {', '.join(forbidden_keys)} "
                "as they are dynamically determined."
            )

        if not isinstance(run_input, list):
            run_input_list = [run_input]
            is_single_input = True
        else:
            run_input_list = run_input
            is_single_input = False

        noise_model: Optional[Union[NoiseModel, str]] = kwargs.get("noise_model")

        if noise_model:
            kwargs["noise_model"] = self._resolve_noise_model(noise_model)

        params: Optional[Union[QuboSolveParams, dict]] = kwargs.get("params")

        if params:
            kwargs["params"] = self._resolve_qubo_params(params)

        jobs: list[qbraid.runtime.QbraidJob] = []
        native_target = self._all_target_specs_native(self._target_spec)
        transpile_option = self._target_spec is not None and self._options.get("transpile") is True

        for i, program in enumerate(run_input_list):
            aux_payload = {}
            program_spec = None
            if transpile_option or not native_target:
                program_alias = get_program_type_alias(program, safe=True)
                program_spec = ProgramSpec(type(program), alias=program_alias)
            if not native_target:
                aux_payload = self._construct_aux_payload(program, program_spec)
            if transpile_option:
                program = self.transpile(program, program_spec)
            is_batched_output = is_single_input and isinstance(program, list)
            program_batch = program if is_batched_output else [program]
            self.validate(program_batch, suppress_device_warning=i != 0)
            for program in program_batch:
                if native_target:
                    aux_payload = self._construct_aux_payload(program, program_spec)
                run_input_json = self.prepare(program)
                self._validate_run_input_payload(run_input_json, self._target_spec)
                runtime_payload = {**aux_payload, **run_input_json}
                job = self.submit(run_input=runtime_payload, shots=shots, tags=tags, **kwargs)
                jobs.append(job)
        return jobs[0] if (len(jobs) == 1 and is_single_input) else jobs

    def estimate_cost(
        self, shots: Optional[int], execution_time: Optional[Union[float, int]]
    ) -> Credits:
        """Estimate the cost of running a quantum job on this device in qBraid credits,
        where 1 credit equals $0.01 USD.

        The estimated cost is based on the device's pricing model, which may include charges per
        task, per shot, and/or per minute. *Note*: The actual price charged may differ from this
        calculation. Visit https://docs.qbraid.com/home/pricing for the latest pricing information
        and details about qBraid credits.

        Args:
            shots (int, optional): The number of quantum circuit executions in the quantum job.
            execution_time (Union[float, int], optional): The estimated time (in minutes) to
                complete the quantum job.

        Returns:
            Credits: The estimated cost for the quantum job in qBraid credits.

        Raises:
            ValueError: If `shots` and `execution_time` are None or 0, or if either is negative.
            QbraidRuntimeError: If unable to retrieve the cost estimate from the qBraid API.
        """
        if not shots:
            shots = None
        if not execution_time:
            execution_time = None

        if shots is None and execution_time is None:
            raise ValueError(
                "At least one of 'shots' or 'execution_time' must be provided to estimate cost."
            )

        if shots is not None:
            if not isinstance(shots, int) or shots < 0:
                raise ValueError("'shots' must be a non-negative integer.")

        if execution_time is not None:
            if not isinstance(execution_time, (int, float)) or execution_time < 0:
                raise ValueError("'execution_time' must be a non-negative number.")

        try:
            cost = self.client.estimate_cost(self.id, shots, execution_time)
        except QuantumServiceRequestError as err:
            if self.profile.provider_name == "IonQ" and not self.profile.simulator:
                raise NotImplementedError(
                    "Cost estimation is not yet supported for IonQ QPUs."
                ) from err
            raise QbraidRuntimeError(
                "Failed to estimate cost due to a service request error."
            ) from err

        return Credits(cost)
